"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Finance functionalities.
"""
__all__ = ['net_present_value', 'income_group', 'gdp']

import shutil
import logging
import warnings
import zipfile

from pathlib import Path

import requests
import numpy as np
import pandas as pd
from pandas_datareader import wb
from cartopy.io import shapereader

from climada.util.files_handler import download_file
from climada.util.constants import SYSTEM_DIR


LOGGER = logging.getLogger(__name__)

WORLD_BANK_WEALTH_ACC = \
"https://databank.worldbank.org/data/download/Wealth-Accounts_CSV.zip"
"""Wealth historical data (1995, 2000, 2005, 2010, 2014) from World Bank (ZIP).
    https://datacatalog.worldbank.org/dataset/wealth-accounting
    Includes variable Produced Capital (NW.PCA.TO)"""

FILE_WORLD_BANK_WEALTH_ACC = "Wealth-AccountsData.csv"

WORLD_BANK_INC_GRP = \
"http://databank.worldbank.org/data/download/site-content/OGHIST.xls"
"""Income group historical data from World bank."""

INCOME_GRP_WB_TABLE = {'L': 1,  # low income
                       'LM': 2,  # lower middle income
                       'UM': 3,  # upper middle income
                       'H': 4,  # high income
                       '..': np.nan  # no data
                      }
"""Meaning of values of world banks' historical table on income groups."""

INCOME_GRP_NE_TABLE = {5: 1,  # Low income
                       4: 2,  # Lower middle income
                       3: 3,  # Upper middle income
                       2: 4,  # High income: nonOECD
                       1: 4  # High income: OECD
                      }
"""Meaning of values of natural earth's income groups."""

FILE_GWP_WEALTH2GDP_FACTORS = 'WEALTH2GDP_factors_CRI_2016.csv'
"""File with wealth-to-GDP factors from the
Credit Suisse's Global Wealth Report 2017 (household wealth)"""

def _nat_earth_shp(resolution='10m', category='cultural',
                   name='admin_0_countries'):
    shp_file = shapereader.natural_earth(resolution=resolution,
                                         category=category, name=name)
    return shapereader.Reader(shp_file)

def net_present_value(years, disc_rates, val_years):
    """Compute net present value.

    Parameters
    ----------
    years : np.array
        array with the sequence of years to consider.
    disc_rates : np.array
        discount rate for every year in years.
    val_years : np.array
        chash flow at each year.

    Returns
    -------
    float
    """
    if years.size != disc_rates.size or years.size != val_years.size:
        raise ValueError(f'Wrong input sizes {years.size}, {disc_rates.size}, {val_years.size}.')

    npv = val_years[-1]
    for val, disc in zip(val_years[-2::-1], disc_rates[-2::-1]):
        npv = val + npv / (1 + disc)

    return npv

def income_group(cntry_iso, ref_year, shp_file=None):
    """Get country's income group from World Bank's data at a given year,
    or closest year value. If no data, get the natural earth's approximation.

    Parameters
    ----------
    cntry_iso : str
        key = ISO alpha_3 country
    ref_year : int
        reference year
    shp_file : cartopy.io.shapereader.Reader, optional
        shape file with
        INCOME_GRP attribute for every country. Load Natural Earth admin0
        if not provided.
    """
    try:
        close_year, close_val = world_bank(cntry_iso, ref_year, 'INC_GRP')
    except (KeyError, IndexError):
        # take value from natural earth repository
        close_year, close_val = nat_earth_adm0(cntry_iso, 'INCOME_GRP',
                                               shp_file=shp_file)
    finally:
        LOGGER.info('Income group %s %s: %s.', cntry_iso, close_year, close_val)

    return close_year, close_val

def gdp(cntry_iso, ref_year, shp_file=None, per_capita=False):
    """Get country's (current value) GDP from World Bank's data at a given year, or
    closest year value. If no data, get the natural earth's approximation.

    Parameters
    ----------
    cntry_iso : str
        key = ISO alpha_3 country
    ref_year : int
        reference year
    shp_file : cartopy.io.shapereader.Reader, optional
        shape file with
        INCOME_GRP attribute for every country. Load Natural Earth admin0
        if not provided.
    per_capita : boolean, optional
        If True, GDP is returned per capita

    Returns
    -------
    float
    """
    if cntry_iso == 'TWN':
        LOGGER.warning('GDP data for TWN is not provided by World Bank. \
                       Instead, IMF data is returned here.')
        close_year, close_val = _gdp_twn(ref_year, per_capita=per_capita)
        return close_year, close_val
    try:
        if per_capita:
            close_year, close_val = world_bank(cntry_iso, ref_year, 'NY.GDP.PCAP.CD')
        else:
            close_year, close_val = world_bank(cntry_iso, ref_year, 'NY.GDP.MKTP.CD')
    except (ValueError, IndexError, requests.exceptions.ConnectionError) as err:
        if isinstance(err, requests.exceptions.ConnectionError):
            LOGGER.warning('Internet connection failed while retrieving GDPs.')
        close_year, close_val = nat_earth_adm0(cntry_iso, 'GDP_MD', 'GDP_YEAR', shp_file)
    LOGGER.info("GDP {} {:d}: {:.3e}.".format(cntry_iso, close_year, close_val))

    return close_year, close_val

def world_bank(cntry_iso, ref_year, info_ind):
    """Get country's GDP from World Bank's data at a given year, or
    closest year value. If no data, get the natural earth's approximation.

    Parameters
    ----------
    cntry_iso : str
        key = ISO alpha_3 country
    ref_year : int
        reference year
    info_ind : str
        indicator of World Bank, e.g. 'NY.GDP.MKTP.CD'. If
        'INC_GRP', historical income groups from excel file used.

    Returns
    -------
    int, float

    Raises
    ------
    IOError, KeyError, IndexError
    """
    if info_ind != 'INC_GRP':
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cntry_gdp = wb.download(indicator=info_ind, country=cntry_iso, start=1960, end=2030)
        years = np.array([int(year) for year in cntry_gdp.index.get_level_values('year')])
        sort_years = np.abs(years - ref_year).argsort()
        close_val = cntry_gdp.iloc[sort_years].dropna()
        close_year = int(close_val.iloc[0].name[1])
        close_val = float(close_val.iloc[0].values)
    else:  # income group level
        fn_ig = SYSTEM_DIR.joinpath('OGHIST.xls')
        dfr_wb = pd.DataFrame()
        try:
            if not fn_ig.is_file():
                file_down = download_file(WORLD_BANK_INC_GRP)
                shutil.move(file_down, fn_ig)
            dfr_wb = pd.read_excel(fn_ig, 'Country Analytical History', skiprows=5)
            dfr_wb = dfr_wb.drop(dfr_wb.index[0:5]).set_index('Unnamed: 0')
            dfr_wb = dfr_wb.replace(INCOME_GRP_WB_TABLE.keys(),
                                    INCOME_GRP_WB_TABLE.values())
        except (IOError, requests.exceptions.ConnectionError) as err:
            raise type(err)('Internet connection failed while downloading '
                            'historical income groups: ' + str(err)) from err

        cntry_dfr = dfr_wb.loc[cntry_iso]
        close_val = cntry_dfr.iloc[np.abs(
            np.array(cntry_dfr.index[1:]) - ref_year).argsort() + 1].dropna()
        close_year = close_val.index[0]
        close_val = int(close_val.iloc[0])

    return close_year, close_val

def nat_earth_adm0(cntry_iso, info_name, year_name=None, shp_file=None):
    """Get country's parameter from natural earth's admin0 shape file.

    Parameters
    ----------
    cntry_iso : str
        key = ISO alpha_3 country
    info_name : str
        attribute to get, e.g. 'GDP_MD', 'INCOME_GRP'.
    year_name : str, optional
        year name of the info_name in shape file,
        e.g. 'GDP_YEAR'
    shp_file : cartopy.io.shapereader.Reader, optional
        shape file with
        INCOME_GRP attribute for every country. Load Natural Earth admin0
        if not provided.

    Returns
    -------
    int, float

    Raises
    ------
    ValueError
    """
    if not shp_file:
        shp_file = _nat_earth_shp('10m', 'cultural', 'admin_0_countries')

    close_val = 0
    close_year = 0
    for info in shp_file.records():
        if info.attributes['ADM0_A3'] == cntry_iso:
            close_val = info.attributes[info_name]
            if year_name:
                close_year = int(info.attributes[year_name])
            break

    if not close_val:
        raise ValueError("No GDP for country %s found." % cntry_iso)

    # the variable name changed in Natural Earth v5.0.0
    if info_name in ['GDP_MD', 'GDP_MD_EST']:
        close_val *= 1e6
    elif info_name == 'INCOME_GRP':
        close_val = INCOME_GRP_NE_TABLE.get(int(close_val[0]))

    return close_year, close_val

def wealth2gdp(cntry_iso, non_financial=True, ref_year=2016,
               file_name=FILE_GWP_WEALTH2GDP_FACTORS):
    """Get country's wealth-to-GDP factor from the
        Credit Suisse's Global Wealth Report 2017 (household wealth).
        Missing value: returns NaN.

    Parameters
    ----------
    cntry_iso : str
        key = ISO alpha_3 country
    non_financial : boolean
        use non-financial wealth (True)
        use total wealth (False)
    ref_year : int
        reference year

    Returns
    -------
    float
    """
    fname = SYSTEM_DIR.joinpath(file_name)
    factors_all_countries = pd.read_csv(fname, sep=',', index_col=None,
                                        header=0, encoding='ISO-8859-1')
    if ref_year != 2016:
        LOGGER.warning('Reference year for the factor to convert GDP to '
                       'wealth was set to 2016 because other years have not '
                       'been implemented yet.')
        ref_year = 2016
    if non_financial:
        try:
            val = factors_all_countries[
                factors_all_countries.country_iso3 == cntry_iso]['NFW-to-GDP-ratio'].values[0]
        except (AttributeError, KeyError, IndexError):
            LOGGER.warning('No data for country, using mean factor.')
            val = factors_all_countries["NFW-to-GDP-ratio"].mean()
    else:
        try:
            val = factors_all_countries[
                factors_all_countries.country_iso3 == cntry_iso]['TW-to-GDP-ratio'].values[0]
        except (AttributeError, KeyError, IndexError):
            LOGGER.warning('No data for country, using mean factor.')
            val = factors_all_countries["TW-to-GDP-ratio"].mean()
    val = np.around(val, 5)
    return ref_year, val

def world_bank_wealth_account(cntry_iso, ref_year, variable_name="NW.PCA.TO",
                              no_land=True):
    """
    Download and unzip wealth accounting historical data (1995, 2000, 2005, 2010, 2014)
    from World Bank (https://datacatalog.worldbank.org/dataset/wealth-accounting).
    Return requested variable for a country (cntry_iso) and a year (ref_year).

    Parameters
    ----------
    cntry_iso : str
        ISO3-code of country, i.e. "CHN" for China
    ref_year : int
        reference year
        - available in data: 1995, 2000, 2005, 2010, 2014
        - other years between 1995 and 2014 are interpolated
        - for years outside range, indicator is scaled
          proportionally to GDP
    variable_name : str
        select one variable, i.e.:
        'NW.PCA.TO': Produced capital stock of country
            incl. manufactured or built assets such as machinery,
            equipment, and physical structures
            and value of built-up urban land (24% mark-up)
        'NW.PCA.PC': Produced capital stock per capita
            incl. manufactured or built assets such as machinery,
            equipment, and physical structures
            and value of built-up urban land (24% mark-up)
        'NW.NCA.TO': Total natural capital of country. Natural capital
            includes the valuation of fossil fuel energy (oil, gas,
            hard and soft coal) and minerals (bauxite, copper, gold,
            iron ore, lead, nickel, phosphate, silver, tin, and zinc),
            agricultural land (cropland and pastureland),
            forests (timber and some nontimber forest products), and
            protected areas.
        'NW.TOW.TO': Total wealth of country.
        Note: Values are measured at market exchange rates in constant 2014 US dollars,
            using a country-specific GDP deflator.
    no_land : boolean
        If True, return produced capital without built-up land value
        (applies to 'NW.PCA.*' only). Default: True.
    """
    try:
        data_file = SYSTEM_DIR.joinpath(FILE_WORLD_BANK_WEALTH_ACC)
        if not data_file.is_file():
            data_file = SYSTEM_DIR.joinpath('Wealth-Accounts_CSV', FILE_WORLD_BANK_WEALTH_ACC)
        if not data_file.is_file():
            if not SYSTEM_DIR.joinpath('Wealth-Accounts_CSV').is_dir():
                SYSTEM_DIR.joinpath('Wealth-Accounts_CSV').mkdir()
            file_down = download_file(WORLD_BANK_WEALTH_ACC)
            zip_ref = zipfile.ZipFile(file_down, 'r')
            zip_ref.extractall(SYSTEM_DIR.joinpath('Wealth-Accounts_CSV'))
            zip_ref.close()
            Path(file_down).unlink()
            LOGGER.debug('Download and unzip complete. Unzipping %s', str(data_file))

        data_wealth = pd.read_csv(data_file, sep=',', index_col=None, header=0)
    except Exception as err:
        raise type(err)(
            'Downloading World Bank Wealth Accounting Data failed: ' + str(err)) from err

    data_wealth = data_wealth[data_wealth['Country Code'].str.contains(cntry_iso)
                              & data_wealth['Indicator Code'].str.contains(variable_name)
                             ].loc[:, '1995':'2014']
    years = list(map(int, list(data_wealth)))
    if data_wealth.size == 0 and 'NW.PCA.TO' in variable_name:  # if country is not found in data
        LOGGER.warning('No data available for country. Using non-financial wealth instead')
        gdp_year, gdp_val = gdp(cntry_iso, ref_year)
        fac = wealth2gdp(cntry_iso)[1]
        return gdp_year, np.around((fac * gdp_val), 1), 0
    if ref_year in years:  # indicator for reference year is available directly
        result = data_wealth.loc[:, str(ref_year)].values[0]
    elif ref_year > np.min(years) and ref_year < np.max(years):  # interpolate
        result = np.interp(ref_year, years, data_wealth.values[0, :])
    elif ref_year < np.min(years):  # scale proportionally to GDP
        gdp_year, gdp0_val = gdp(cntry_iso, np.min(years))
        gdp_year, gdp_val = gdp(cntry_iso, ref_year)
        result = data_wealth.values[0, 0] * gdp_val / gdp0_val
        ref_year = gdp_year
    else:
        gdp_year, gdp0_val = gdp(cntry_iso, np.max(years))
        gdp_year, gdp_val = gdp(cntry_iso, ref_year)
        result = data_wealth.values[0, -1] * gdp_val / gdp0_val
        ref_year = gdp_year
    if 'NW.PCA.' in variable_name and no_land:
        # remove value of built-up land from produced capital
        result = result / 1.24
    return ref_year, np.around(result, 1), 1

def _gdp_twn(ref_year, per_capita=False):
    """returns GDP for TWN (Republic of China / Taiwan Province of China) based
    on a CSV sheet downloaded from the
    International Monetary Fund (IMF).
    The reason for this special treatment is the
    lack of GDP data for TWN in the World Bank data

    Data Source:
        https://www.imf.org/external/pubs/ft/weo/2019/02/weodata/index.aspx
        https://www.imf.org/external/pubs/ft/weo/2019/02/weodata/weorept.aspx?sy=1980&ey=2024&scsm=1&ssd=1&sic=1&sort=country&ds=.&br=1&pr1.x=42&pr1.y=10&c=528&s=NGDPD%2CNGDP_D%2CNGDPDPC&grp=0&a=
        (saved as CSV with name GDP_TWN_IMF_WEO_data in SYSTEM_DIR)

    Parameters
    ----------
    ref_year : int
        reference year, i.e. the year for which a GDP value is required
    per_capita : boolean
        return GDP per capita? Default False.

    Returns
    -------
    float
    """
    fname = 'GDP_TWN_IMF_WEO_data.csv'
    if not SYSTEM_DIR.joinpath(fname).is_file():
        raise FileNotFoundError(f'File {fname} not found in SYSTEM_DIR')
    if per_capita:
        var_name = 'Gross domestic product per capita, current prices'
    else:
        var_name = 'Gross domestic product, current prices'
    if ref_year < 1980:
        close_year = 1980
    elif ref_year > 2024:
        close_year = 2024
    else:
        close_year = ref_year
    data = pd.read_csv(SYSTEM_DIR.joinpath('GDP_TWN_IMF_WEO_data.csv'),
                       index_col=None, header=0)
    close_val = data.loc[data['Subject Descriptor'] == var_name, str(close_year)].values[0]
    close_val = float(close_val.replace(',', ''))
    if not per_capita:
        close_val = close_val * 1e9
    return close_year, close_val



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test coordinates module.
"""

import unittest
from pathlib import Path

from cartopy.io import shapereader
import geopandas as gpd
import numpy as np
from pyproj.crs import CRS as PCRS
import shapely
from shapely.geometry import box, Point
from rasterio.windows import Window
from rasterio.warp import Resampling
from rasterio import Affine
from rasterio.crs import CRS as RCRS
import rasterio.transform

from climada import CONFIG
from climada.util.constants import HAZ_DEMO_FL, DEF_CRS, ONE_LAT_KM
import climada.util.coordinates as u_coord

DATA_DIR = CONFIG.util.test_data.dir()

def data_arrays_resampling_demo():
    """init demo data arrays (2d) and meta data for resampling"""
    data_arrays = [
        # demo pop:
        np.array([[0, 1, 2], [3, 4, 5]], dtype='float32'),
        np.array([[0, 1, 2], [3, 4, 5]], dtype='float32'),
        # demo nightlight:
        np.array([[2, 10, 0, 0, 0, 0],
                  [10, 2, 10, 0, 0, 0],
                  [0, 0, 0, 0, 1, 1],
                  [1, 0, 0, 0, 1, 1]], dtype='float32'),
    ]

    meta_list = [{'driver': 'GTiff',
                  'dtype': 'float32',
                  'nodata': -3.4028230607370965e+38,
                  'width': 3,
                  'height': 2,
                  'count': 1,
                  'crs': RCRS.from_epsg(4326),
                  'transform': Affine(1, 0.0, -10, 0.0, -1, 40),
                  },
                 {'driver': 'GTiff',
                  'dtype': 'float32',
                  'nodata': -3.4028230607370965e+38,
                  'width': 3,
                  'height': 2,
                  'count': 1,
                  'crs': RCRS.from_epsg(4326),
                  # shifted by 1 degree latitude to the north:
                  'transform': Affine(1, 0.0, -10, 0.0, -1, 41),
                  },
                 {'driver': 'GTiff',
                  'dtype': 'float32',
                  'nodata': None,
                  'width': 6,
                  'height': 4,
                  'count': 1,
                  'crs': RCRS.from_epsg(4326),
                  # higher resolution:
                  'transform': Affine(.5, 0.0, -10, 0.0, -.5, 40),
                  }]
    return data_arrays, meta_list

class TestFunc(unittest.TestCase):
    """Test auxiliary functions"""
    def test_lon_normalize(self):
        """Test the longitude normalization function"""
        data = np.array([-180, 20.1, -30, 190, -350])

        # test in place operation
        u_coord.lon_normalize(data)
        np.testing.assert_array_almost_equal(data, [180, 20.1, -30, -170, 10])

        # test with specific center and return value
        data = u_coord.lon_normalize(data, center=-170)
        np.testing.assert_array_almost_equal(data, [-180, -339.9, -30, -170, 10])

        # test with center determined automatically (which is 280.05)
        data = u_coord.lon_normalize(data, center=None)
        np.testing.assert_array_almost_equal(data, [180, 380.1, 330, 190, 370])

    def test_latlon_bounds(self):
        """Test latlon_bounds function"""
        lat, lon = np.array([0, -2, 5]), np.array([-179, 175, 178])
        bounds = u_coord.latlon_bounds(lat, lon)
        self.assertEqual(bounds, (175, -2, 181, 5))
        bounds = u_coord.latlon_bounds(lat, lon, buffer=1)
        self.assertEqual(bounds, (174, -3, 182, 6))

        # buffer exceeding antimeridian
        lat, lon = np.array([0, -2.1, 5]), np.array([-179.5, -175, -178])
        bounds = u_coord.latlon_bounds(lat, lon, buffer=1)
        self.assertEqual(bounds, (179.5, -3.1, 186, 6))

        # longitude values need to be normalized before they lie between computed bounds:
        lon_mid = 0.5 * (bounds[0] + bounds[2])
        lon = u_coord.lon_normalize(lon, center=lon_mid)
        self.assertTrue(np.all((bounds[0] <= lon) & (lon <= bounds[2])))

        # data covering almost the whole longitudinal range
        lat, lon = np.linspace(-90, 90, 180), np.linspace(-180.0, 179, 360)
        bounds = u_coord.latlon_bounds(lat, lon)
        self.assertEqual(bounds, (-179, -90, 180, 90))
        bounds = u_coord.latlon_bounds(lat, lon, buffer=1)
        self.assertEqual(bounds, (-180, -90, 180, 90))

    def test_geosph_vector(self):
        """Test conversion from lat/lon to unit vector on geosphere"""
        data = np.array([[0, 0], [-13, 179]], dtype=np.float64)
        vn, vbasis = u_coord.latlon_to_geosph_vector(data[:, 0], data[:, 1], basis=True)
        basis_scal = (vbasis[..., 0, :] * vbasis[..., 1, :]).sum(axis=-1)
        basis_norm = np.linalg.norm(vbasis, axis=-1)
        self.assertTrue(np.allclose(np.linalg.norm(vn, axis=-1), 1))
        self.assertTrue(np.allclose(basis_scal, 0))
        self.assertTrue(np.allclose(basis_norm, 1))

    def test_dist_approx_pass(self):
        """Test approximate distance functions"""
        data = np.array([
            # lat1, lon1, lat2, lon2, dist, dist_sph
            [45.5, -32.1, 14, 56, 7702.88906574, 8750.64119051],
            [45.5, 147.8, 14, -124, 7709.82781473, 8758.34146833],
            [45.5, 507.9, 14, -124, 7702.88906574, 8750.64119051],
            [45.5, -212.2, 14, -124, 7709.82781473, 8758.34146833],
            [-3, -130.1, 4, -30.5, 11079.7217421, 11087.0352544],
        ])
        compute_dist = np.stack([
            u_coord.dist_approx(data[:, None, 0], data[:, None, 1],
                                data[:, None, 2], data[:, None, 3],
                                method="equirect")[:, 0, 0],
            u_coord.dist_approx(data[:, None, 0], data[:, None, 1],
                                data[:, None, 2], data[:, None, 3],
                                method="geosphere")[:, 0, 0],
        ], axis=-1)
        self.assertEqual(compute_dist.shape[0], data.shape[0])
        for d, cd in zip(data[:, 4:], compute_dist):
            self.assertAlmostEqual(d[0], cd[0])
            self.assertAlmostEqual(d[1], cd[1])

        for units, factor in zip(["radian", "degree", "km"],
                                 [np.radians(1.0), 1.0, u_coord.ONE_LAT_KM]):
            factor /= u_coord.ONE_LAT_KM
            compute_dist = np.stack([
                u_coord.dist_approx(data[:, None, 0], data[:, None, 1],
                                    data[:, None, 2], data[:, None, 3],
                                    method="equirect", units=units)[:, 0, 0],
                u_coord.dist_approx(data[:, None, 0], data[:, None, 1],
                                    data[:, None, 2], data[:, None, 3],
                                    method="geosphere", units=units)[:, 0, 0],
            ], axis=-1)
            self.assertEqual(compute_dist.shape[0], data.shape[0])
            for d, cd in zip(data[:, 4:], compute_dist):
                self.assertAlmostEqual(d[0] * factor, cd[0])
                self.assertAlmostEqual(d[1] * factor, cd[1])

    def test_dist_approx_log_pass(self):
        """Test log-functionality of approximate distance functions"""
        data = np.array([
            # lat1, lon1, lat2, lon2, dist, dist_sph
            [0, 0, 0, 1, 111.12, 111.12],
            [-13, 179, 5, -179, 2011.84774049, 2012.30698122],
        ])
        for i, method in enumerate(["equirect", "geosphere"]):
            for units, factor in zip(["radian", "degree", "km"],
                                     [np.radians(1.0), 1.0, u_coord.ONE_LAT_KM]):
                factor /= u_coord.ONE_LAT_KM
                dist, vec = u_coord.dist_approx(data[:, None, 0], data[:, None, 1],
                                                data[:, None, 2], data[:, None, 3],
                                                log=True, method=method, units=units)
                dist, vec = dist[:, 0, 0], vec[:, 0, 0]
                np.testing.assert_allclose(np.linalg.norm(vec, axis=-1), dist)
                np.testing.assert_allclose(dist, data[:, 4 + i] * factor)
                # both points on equator (no change in latitude)
                self.assertAlmostEqual(vec[0, 0], 0)
                # longitude from 179 to -179 is positive (!) in lon-direction
                np.testing.assert_array_less(100, vec[1, :] / factor)


    def test_get_gridcellarea(self):
        """Test get_gridcellarea function to calculate the gridcellarea from a given latitude"""

        lat = np.array([54.75, 54.25])
        resolution = 0.5
        area = u_coord.get_gridcellarea(lat, resolution)

        self.assertAlmostEqual(area[0], 178159.73363005)
        self.assertAlmostEqual(area[1], 180352.82386516)
        self.assertEqual(lat.shape, area.shape)

    def test_read_vector_pass(self):
        """Test one columns data"""
        shp_file = shapereader.natural_earth(resolution='110m', category='cultural',
                                             name='populated_places_simple')
        lat, lon, geometry, intensity = u_coord.read_vector(shp_file, ['pop_min', 'pop_max'])

        self.assertTrue(u_coord.equal_crs(geometry.crs, u_coord.NE_EPSG))
        self.assertEqual(geometry.size, lat.size)
        self.assertAlmostEqual(lon[0], 12.453386544971766)
        self.assertAlmostEqual(lon[-1], 114.18306345846304)
        self.assertAlmostEqual(lat[0], 41.903282179960115)
        self.assertAlmostEqual(lat[-1], 22.30692675357551)

        self.assertEqual(intensity.shape, (2, 243))
        # population min
        self.assertEqual(intensity[0, 0], 832)
        self.assertEqual(intensity[0, -1], 4551579)
        # population max
        self.assertEqual(intensity[1, 0], 832)
        self.assertEqual(intensity[1, -1], 7206000)

    def test_compare_crs(self):
        """Compare two crs"""
        crs_one = 'epsg:4326'
        crs_two = {'init': 'epsg:4326', 'no_defs': True}
        self.assertTrue(u_coord.equal_crs(crs_one, crs_two))

    def test_set_df_geometry_points_pass(self):
        """Test set_df_geometry_points"""
        df_val = gpd.GeoDataFrame()
        df_val['latitude'] = np.ones(10) * 40.0
        df_val['longitude'] = np.ones(10) * 0.50

        u_coord.set_df_geometry_points(df_val, crs='epsg:2202')
        self.assertTrue(np.allclose(df_val.geometry[:].x.values, np.ones(10) * 0.5))
        self.assertTrue(np.allclose(df_val.geometry[:].y.values, np.ones(10) * 40.))

    def test_convert_wgs_to_utm_pass(self):
        """Test convert_wgs_to_utm"""
        lat, lon = 17.346597, -62.768669
        epsg = u_coord.convert_wgs_to_utm(lon, lat)
        self.assertEqual(epsg, 32620)

        lat, lon = 41.522410, 1.891026
        epsg = u_coord.convert_wgs_to_utm(lon, lat)
        self.assertEqual(epsg, 32631)

    def test_to_crs_user_input(self):
        pcrs = PCRS.from_epsg(4326)
        rcrs = RCRS.from_epsg(4326)

        # are they the default?
        self.assertEqual(pcrs, PCRS.from_user_input(u_coord.to_crs_user_input(DEF_CRS)))
        self.assertEqual(rcrs, RCRS.from_user_input(u_coord.to_crs_user_input(DEF_CRS)))

        # can they be understood from the provider?
        for arg in ['epsg:4326', b'epsg:4326', DEF_CRS, 4326,
                    {'init': 'epsg:4326', 'no_defs': True},
                    b'{"init": "epsg:4326", "no_defs": True}']:
            self.assertEqual(pcrs, PCRS.from_user_input(u_coord.to_crs_user_input(arg)))
            self.assertEqual(rcrs, RCRS.from_user_input(u_coord.to_crs_user_input(arg)))

        # are they noticed?
        for arg in [4326.0, [4326]]:
            with self.assertRaises(ValueError):
                u_coord.to_crs_user_input(arg)
        with self.assertRaises(SyntaxError):
            u_coord.to_crs_user_input('{init: epsg:4326, no_defs: True}')

    def test_country_to_iso(self):
        name_list = [
            '', 'United States', 'Argentina', 'Japan', 'Australia', 'Norway', 'Madagascar']
        al2_list = ['', 'US', 'AR', 'JP', 'AU', 'NO', 'MG']
        al3_list = ['', 'USA', 'ARG', 'JPN', 'AUS', 'NOR', 'MDG']
        num_list = [0, 840, 32, 392, 36, 578, 450]
        natid_list = [0, 217, 9, 104, 13, 154, 128]

        # examples from docstring:
        self.assertEqual(u_coord.country_to_iso(840), "USA")
        self.assertEqual(u_coord.country_to_iso("United States", representation="alpha2"), "US")
        self.assertEqual(u_coord.country_to_iso(["United States of America", "SU"], "numeric"),
                         [840, 810])
        self.assertEqual(u_coord.country_to_iso(["XK", "Dhekelia"], "numeric"), [983, 907])

        # test cases:
        iso_lists = [name_list, al2_list, al3_list, num_list]
        for l1 in iso_lists:
            for l2, representation in zip(iso_lists, ["name", "alpha2", "alpha3", "numeric"]):
                self.assertEqual(u_coord.country_to_iso(l1, representation), l2)

        # deprecated API `country_iso_alpha2numeric`
        self.assertEqual(u_coord.country_iso_alpha2numeric(al3_list[-2]), num_list[-2])
        self.assertEqual(u_coord.country_iso_alpha2numeric(al3_list), num_list)

        # test fillvalue-parameter for unknown countries
        self.assertEqual(u_coord.country_to_iso("ABC", fillvalue="XOT"), "XOT")

        # conversion to/from ISIMIP natid
        self.assertEqual(u_coord.country_iso2natid(al3_list[2]), natid_list[2])
        self.assertEqual(u_coord.country_iso2natid(al3_list), natid_list)
        self.assertEqual(u_coord.country_natid2iso(natid_list, "numeric"), num_list)
        self.assertEqual(u_coord.country_natid2iso(natid_list, "alpha2"), al2_list)
        self.assertEqual(u_coord.country_natid2iso(natid_list, "alpha3"), al3_list)
        self.assertEqual(u_coord.country_natid2iso(natid_list), al3_list)
        self.assertEqual(u_coord.country_natid2iso(natid_list[1]), al3_list[1])

class TestAssign(unittest.TestCase):
    """Test coordinate assignment functions"""

    def test_assign_grid_points(self):
        """Test assign_grid_points function"""
        res = (1, -0.5)
        pt_bounds = (5, 40, 10, 50)
        height, width, transform = u_coord.pts_to_raster_meta(pt_bounds, res)
        xy = np.array([[5, 50], [6, 50], [5, 49.5], [10, 40]])
        out = u_coord.assign_grid_points(xy[:, 0], xy[:, 1], width, height, transform)
        np.testing.assert_array_equal(out, [0, 1, 6, 125])

        res = (0.5, -0.5)
        height, width, transform = u_coord.pts_to_raster_meta(pt_bounds, res)
        xy = np.array([[5.1, 49.95], [5.99, 50.05], [5, 49.6], [10.1, 39.8]])
        out = u_coord.assign_grid_points(xy[:, 0], xy[:, 1], width, height, transform)
        np.testing.assert_array_equal(out, [0, 2, 11, 230])

        res = (1, -1)
        pt_bounds = (-20, -40, -10, -30)
        height, width, transform = u_coord.pts_to_raster_meta(pt_bounds, res)
        xy = np.array([[-10, -40], [-30, -40]])
        out = u_coord.assign_grid_points(xy[:, 0], xy[:, 1], width, height, transform)
        np.testing.assert_array_equal(out, [120, -1])

    def test_dist_sqr_approx_pass(self):
        """Test approximate distance helper function."""
        lats1 = 45.5
        lons1 = -32.2
        cos_lats1 = np.cos(np.radians(lats1))
        lats2 = 14
        lons2 = 56
        self.assertAlmostEqual(
            7709.827814738594,
            np.sqrt(u_coord._dist_sqr_approx(lats1, lons1, cos_lats1, lats2, lons2)) * ONE_LAT_KM)

    def test_wrong_distance_fail(self):
        """Check exception is thrown when wrong distance is given"""
        with self.assertRaises(ValueError) as cm:
            u_coord.assign_coordinates(np.ones((10, 2)), np.ones((7, 2)), distance='distance')
        self.assertIn('Coordinate assignment with "distance" distance is not supported.',
                      str(cm.exception))

    def data_input_values(self):
        """Default input coordinates and centroids values"""
        # Load exposures coordinates from demo entity file
        exposures = np.array([
            [26.933899, -80.128799],
            [26.957203, -80.098284],
            [26.783846, -80.748947],
            [26.645524, -80.550704],
            [26.897796, -80.596929],
            [26.925359, -80.220966],
            [26.914768, -80.07466],
            [26.853491, -80.190281],
            [26.845099, -80.083904],
            [26.82651, -80.213493],
            [26.842772, -80.0591],
            [26.825905, -80.630096],
            [26.80465, -80.075301],
            [26.788649, -80.069885],
            [26.704277, -80.656841],
            [26.71005, -80.190085],
            [26.755412, -80.08955],
            [26.678449, -80.041179],
            [26.725649, -80.1324],
            [26.720599, -80.091746],
            [26.71255, -80.068579],
            [26.6649, -80.090698],
            [26.664699, -80.1254],
            [26.663149, -80.151401],
            [26.66875, -80.058749],
            [26.638517, -80.283371],
            [26.59309, -80.206901],
            [26.617449, -80.090649],
            [26.620079, -80.055001],
            [26.596795, -80.128711],
            [26.577049, -80.076435],
            [26.524585, -80.080105],
            [26.524158, -80.06398],
            [26.523737, -80.178973],
            [26.520284, -80.110519],
            [26.547349, -80.057701],
            [26.463399, -80.064251],
            [26.45905, -80.07875],
            [26.45558, -80.139247],
            [26.453699, -80.104316],
            [26.449999, -80.188545],
            [26.397299, -80.21902],
            [26.4084, -80.092391],
            [26.40875, -80.1575],
            [26.379113, -80.102028],
            [26.3809, -80.16885],
            [26.349068, -80.116401],
            [26.346349, -80.08385],
            [26.348015, -80.241305],
            [26.347957, -80.158855]
        ])

        # Define centroids
        centroids = np.zeros((100, 2))
        inext = 0
        for ilon in range(10):
            for ilat in range(10):
                centroids[inext][0] = 20 + ilat + 1
                centroids[inext][1] = -85 + ilon + 1

                inext = inext + 1

        return centroids, exposures

    def data_ref(self):
        """Default output reference"""
        return np.array([46, 46, 36, 36, 36, 46, 46, 46, 46, 46, 46,
                         36, 46, 46, 36, 46, 46, 46, 46, 46, 46, 46,
                         46, 46, 46, 46, 46, 46, 46, 46, 46, 46, 46,
                         46, 46, 46, 45, 45, 45, 45, 45, 45, 45, 45,
                         45, 45, 45, 45, 45, 45])

    def data_ref_40(self):
        """Default output reference for maximum distance threshold 40km"""
        return np.array([46, 46, 36, -1, -1, 46, 46, 46, 46, 46, 46, -1, 46, 46,
                         -1, 46, 46, 46, 46, 46, 46, 46, 46, -1, 46, -1, -1, -1,
                         -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
                         -1, -1, -1, -1, -1, 45, -1, -1])

    def data_antimeridian_values(self):
        """Default input coordinates and centroids value crossing antimerdian"""
        exposures = np.array([
            [0, -179.99],
            [0, 179.99],
            [5, -179.09],
            [-5, 179.09],
            [0, 130],
            [0, -130]
        ])

        # Define centroids
        centroids = np.zeros((100, 2))
        inext = 0
        for ilon in range(10):
            for ilat in range(10):
                centroids[inext][0] = -5 + ilat
                if ilat -5 <= 0:
                    centroids[inext][1] = 170 + ilon + 1
                else:
                    centroids[inext][1] = -170 - ilon
                inext = inext + 1

        return centroids, exposures

    def data_ref_antimeridian(self):
        """Default output reference for maximum distance threshold 100km for
        points crossing the anti-merdiian"""
        return np.array([95, 95, -1, 80, -1, -1])

    def normal_pass(self, dist):
        """Checking result against matlab climada_demo_step_by_step"""
        # Load input
        centroids, exposures = self.data_input_values()

        # Interpolate with default threshold
        neighbors = u_coord.assign_coordinates(exposures, centroids, distance=dist)
        # Reference output
        ref_neighbors = self.data_ref()
        # Check results
        self.assertEqual(exposures.shape[0], len(neighbors))
        np.testing.assert_array_equal(neighbors, ref_neighbors)

    def normal_warning(self, dist):
        """Checking that a warning is raised when minimum distance greater
        than threshold"""
        # Load input
        centroids, exposures = self.data_input_values()

        # Interpolate with lower threshold to raise warnings
        threshold = 40
        with self.assertLogs('climada.util.coordinates', level='INFO') as cm:
            neighbors = u_coord.assign_coordinates(
                exposures, centroids, distance=dist, threshold=threshold)
        self.assertIn("Distance to closest centroid", cm.output[1])

        ref_neighbors = self.data_ref_40()
        np.testing.assert_array_equal(neighbors, ref_neighbors)

    def repeat_coord_pass(self, dist):
        """Check that exposures with the same coordinates have same neighbors"""

        # Load input
        centroids, exposures = self.data_input_values()

        # Repeat a coordinate
        exposures[2, :] = exposures[0, :]

        # Interpolate with default threshold
        neighbors = u_coord.assign_coordinates(exposures, centroids, distance=dist)

        # Check output neighbors have same size as coordinates
        self.assertEqual(len(neighbors), exposures.shape[0])
        # Check copied coordinates have same neighbors
        self.assertEqual(neighbors[2], neighbors[0])

    def antimeridian_warning(self, dist):
        """Check that a warning is raised when minimum distance greater than threshold"""
        # Load input
        centroids, exposures = self.data_antimeridian_values()

        # Interpolate with lower threshold to raise warnings
        threshold = 100
        with self.assertLogs('climada.util.coordinates', level='INFO') as cm:
            neighbors = u_coord.assign_coordinates(
                exposures, centroids, distance=dist, threshold=threshold)
        self.assertIn("Distance to closest centroid", cm.output[1])

        np.testing.assert_array_equal(neighbors, self.data_ref_antimeridian())

    def test_approx_normal_pass(self):
        """Call normal_pass test for approxiamte distance"""
        self.normal_pass('approx')

    def test_approx_normal_warning(self):
        """Call normal_warning test for approxiamte distance"""
        self.normal_warning('approx')

    def test_approx_repeat_coord_pass(self):
        """Call repeat_coord_pass test for approxiamte distance"""
        self.repeat_coord_pass('approx')

    def test_approx_antimeridian_warning(self):
        """Call normal_warning test for approximate distance"""
        self.antimeridian_warning('approx')

    def test_haver_normal_pass(self):
        """Call normal_pass test for haversine distance"""
        self.normal_pass('haversine')

    def test_haver_normal_warning(self):
        """Call normal_warning test for haversine distance"""
        self.normal_warning('haversine')

    def test_haver_repeat_coord_pass(self):
        """Call repeat_coord_pass test for haversine distance"""
        self.repeat_coord_pass('haversine')

    def test_haver_antimeridian_warning(self):
        """Call normal_warning test for haversine distance"""
        self.antimeridian_warning('haversine')

    def test_euc_normal_pass(self):
        """Call normal_pass test for euclidean distance"""
        self.normal_pass('euclidean')

    def test_euc_normal_warning(self):
        """Call normal_warning test for euclidean distance"""
        self.normal_warning('euclidean')

    def test_euc_repeat_coord_pass(self):
        """Call repeat_coord_pass test for euclidean distance"""
        self.repeat_coord_pass('euclidean')

    def test_euc_antimeridian_warning(self):
        """Call normal_warning test for euclidean distance"""
        self.antimeridian_warning('euclidean')

    def test_diff_outcomes(self):
        """Different NN interpolation outcomes"""
        threshold = 100000

        # Define centroids
        lons = np.arange(-160, 180+1, 20)
        lats = np.arange(-60, 60+1, 20)
        lats, lons = [arr.ravel() for arr in np.meshgrid(lats, lons)]
        centroids = np.transpose([lats, lons])

        # Define exposures
        exposures = np.array([
            [49.9, 9],
            [49.5, 9],
            [0, -175]
            ])

        # Neighbors
        ref_neighbors = [
            [62, 62, 3],
            [62, 61, 122],
            [61, 61, 3],
            ]

        dist_list = ['approx', 'haversine', 'euclidean']
        kwargs_list = [
            {'check_antimeridian':False},
            {},
            {'check_antimeridian':False}
            ]

        for dist, ref, kwargs in zip(dist_list, ref_neighbors, kwargs_list):
            neighbors = u_coord.assign_coordinates(
                exposures, centroids, distance=dist, threshold=threshold, **kwargs)
            np.testing.assert_array_equal(neighbors, ref)

    def test_assign_coordinates(self):
        """Test assign_coordinates function"""
        # note that the coordinates are in lat/lon
        coords = np.array([(0.2, 2), (0, 0), (0, 2), (2.1, 3), (1, 1), (-1, 1), (0, 179.9)])
        coords_to_assign = np.array([(2.1, 3), (0, 0), (0, 2), (0.9, 1.0), (0, -179.9)])
        expected_results = [
            # test with different thresholds (in km)
            (100, [2, 1, 2, 0, 3, -1, 4]),
            (20, [-1, 1, 2, 0, 3, -1, -1]),
            (0, [-1, 1, 2, 0, -1, -1, -1]),
        ]

        for distance in ["euclidean", "haversine", "approx"]:
            # make sure that it works for both float32 and float64
            for test_dtype in [np.float64, np.float32]:
                coords_typed = coords.astype(test_dtype)
                coords_to_assign_typed = coords_to_assign.astype(test_dtype)
                for thresh, result in expected_results:
                    assigned_idx = u_coord.assign_coordinates(
                        coords_typed, coords_to_assign_typed,
                        distance=distance, threshold=thresh)
                    np.testing.assert_array_equal(assigned_idx, result)

            #test empty coords_to_assign
            coords_to_assign_empty = np.array([])
            result = [-1, -1, -1, -1, -1, -1, -1]
            assigned_idx = u_coord.assign_coordinates(
                coords, coords_to_assign_empty, distance=distance, threshold=thresh)
            np.testing.assert_array_equal(assigned_idx, result)

            #test empty coords
            coords_empty = np.array([])
            result = np.array([])
            assigned_idx = u_coord.assign_coordinates(
                coords_empty, coords_to_assign, distance=distance, threshold=thresh)
            np.testing.assert_array_equal(assigned_idx, result)

class TestGetGeodata(unittest.TestCase):
    def test_nat_earth_resolution_pass(self):
        """Correct resolution."""
        self.assertEqual(u_coord.nat_earth_resolution(10), '10m')
        self.assertEqual(u_coord.nat_earth_resolution(50), '50m')
        self.assertEqual(u_coord.nat_earth_resolution(110), '110m')

    def test_nat_earth_resolution_fail(self):
        """Wrong resolution."""
        with self.assertRaises(ValueError):
            u_coord.nat_earth_resolution(11)
        with self.assertRaises(ValueError):
            u_coord.nat_earth_resolution(51)
        with self.assertRaises(ValueError):
            u_coord.nat_earth_resolution(111)

    def test_get_coastlines_all_pass(self):
        """Check get_coastlines function over whole earth"""
        coast = u_coord.get_coastlines(resolution=110)
        tot_bounds = coast.total_bounds
        self.assertEqual((134, 1), coast.shape)
        self.assertAlmostEqual(tot_bounds[0], -180)
        self.assertAlmostEqual(tot_bounds[1], -85.60903777)
        self.assertAlmostEqual(tot_bounds[2], 180.00000044)
        self.assertAlmostEqual(tot_bounds[3], 83.64513)

    def test_get_coastlines_pass(self):
        """Check get_coastlines function in defined extent"""
        bounds = (-100, -55, -20, 35)
        coast = u_coord.get_coastlines(bounds, resolution=110)
        ex_box = box(bounds[0], bounds[1], bounds[2], bounds[3])
        self.assertEqual(coast.shape[0], 14)
        self.assertTrue(coast.total_bounds[2] < 0)
        for _row, line in coast.iterrows():
            if not ex_box.intersects(line.geometry):
                self.assertEqual(1, 0)

    def test_get_land_geometry_country_pass(self):
        """get_land_geometry with selected countries."""
        iso_countries = ['DEU', 'VNM']
        res = u_coord.get_land_geometry(iso_countries, 110)
        self.assertIsInstance(res, shapely.geometry.multipolygon.MultiPolygon)
        for res, ref in zip(res.bounds, (5.85248986800, 8.56557851800,
                                         109.47242272200, 55.065334377000)):
            self.assertAlmostEqual(res, ref)

        iso_countries = ['ESP']
        res = u_coord.get_land_geometry(iso_countries, 110)
        self.assertIsInstance(res, shapely.geometry.multipolygon.MultiPolygon)
        for res, ref in zip(res.bounds, (-18.16722571499986, 27.642238674000,
                                         4.337087436000, 43.793443101)):
            self.assertAlmostEqual(res, ref)

        iso_countries = ['FRA']
        res = u_coord.get_land_geometry(iso_countries, 110)
        self.assertIsInstance(res, shapely.geometry.multipolygon.MultiPolygon)
        for res, ref in zip(res.bounds, (-61.79784094999991, -21.37078215899993,
                                         55.854502800000034, 51.08754088371883)):
            self.assertAlmostEqual(res, ref)

    def test_get_land_geometry_extent_pass(self):
        """get_land_geometry with selected countries."""
        lat = np.array([28.203216, 28.555994, 28.860875])
        lon = np.array([-16.567489, -18.554130, -9.532476])
        res = u_coord.get_land_geometry(extent=(np.min(lon), np.max(lon),
                                np.min(lat), np.max(lat)), resolution=10)
        self.assertIsInstance(res, shapely.geometry.multipolygon.MultiPolygon)
        self.assertAlmostEqual(res.bounds[0], -18.002186653)
        self.assertAlmostEqual(res.bounds[1], lat[0])
        self.assertAlmostEqual(res.bounds[2], np.max(lon))
        self.assertAlmostEqual(res.bounds[3], np.max(lat))

    def test_get_land_geometry_all_pass(self):
        """get_land_geometry with all earth."""
        res = u_coord.get_land_geometry(resolution=110)
        self.assertIsInstance(res, shapely.geometry.multipolygon.MultiPolygon)
        self.assertAlmostEqual(res.area, 21496.99098799273)

    def test_on_land_pass(self):
        """check point on land with 1:50.000.000 resolution."""
        lat = np.array([28.203216, 28.555994, 28.860875])
        lon = np.array([-16.567489, -18.554130, -9.532476])
        res = u_coord.coord_on_land(lat, lon)
        self.assertEqual(res.size, 3)
        self.assertTrue(res[0])
        self.assertFalse(res[1])
        self.assertTrue(res[2])

    def test_dist_to_coast(self):
        """Test point in coast and point not in coast"""
        points = np.array([
            # Caribbean Sea:
            [13.208333333333329, -59.625000000000014],
            # South America:
            [-12.497529, -58.849505],
            # Very close to coast of Somalia:
            [1.96768, 45.23219],
        ])
        dists = [2594.2071059573445, 1382985.2459744606, 0.088222234]
        for d, p in zip(dists, points):
            res = u_coord.dist_to_coast(*p)
            self.assertAlmostEqual(d, res[0])

        # All at once requires more than one UTM
        res = u_coord.dist_to_coast(points)
        for d, r in zip(dists, res):
            self.assertAlmostEqual(d, r)

    def test_dist_to_coast_nasa(self):
        """Test point in coast and point not in coast"""
        points = np.array([
            # Caribbean Sea:
            [13.208333333333329, -59.625000000000014],
            # South America:
            [-12.497529, -58.849505],
            # Very close to coast of Somalia:
            [1.96475615, 45.23249055],
        ])
        dists = [-3000, -1393549.5, 48.77]
        dists_lowres = [416.66666667, 1393448.09801077, 1191.38205367]
        # Warning: This will download more than 300 MB of data!
        result = u_coord.dist_to_coast_nasa(points[:, 0], points[:, 1], highres=True, signed=True)
        result_lowres = u_coord.dist_to_coast_nasa(points[:, 0], points[:, 1])
        for d, r in zip(dists, result):
            self.assertAlmostEqual(d, r)
        for d, r in zip(dists_lowres, result_lowres):
            self.assertAlmostEqual(d, r)

    def test_get_country_geometries_country_pass(self):
        """get_country_geometries with selected countries. issues with the
        natural earth data should be caught by test_get_land_geometry_* since
        it's very similar"""
        iso_countries = ['NLD', 'VNM']
        res = u_coord.get_country_geometries(iso_countries, resolution=110)
        self.assertIsInstance(res, gpd.geodataframe.GeoDataFrame)
        self.assertEqual(res.shape[0], 2)

    def test_get_country_geometries_country_norway_pass(self):
        """test correct numeric ISO3 for country Norway"""
        iso_countries = ['NOR']
        extent = [10, 11, 55, 60]
        res1 = u_coord.get_country_geometries(iso_countries)
        res2 = u_coord.get_country_geometries(extent=extent)
        self.assertEqual(res1.ISO_N3.values[0], '578')
        self.assertIn('578', res2.ISO_N3.values)
        self.assertIn('NOR', res2.ISO_A3.values)
        self.assertIn('Denmark', res2.NAME.values)
        self.assertIn('Norway', res2.NAME.values)
        self.assertNotIn('Sweden', res2.NAME.values)

    def test_get_country_geometries_extent_pass(self):
        """get_country_geometries by selecting by extent"""
        lat = np.array([28.203216, 28.555994, 28.860875])
        lon = np.array([-16.567489, -18.554130, -9.532476])

        res = u_coord.get_country_geometries(extent=(
            np.min(lon), np.max(lon),
            np.min(lat), np.max(lat)
        ))

        self.assertIsInstance(res, gpd.geodataframe.GeoDataFrame)
        self.assertTrue(
            np.allclose(res.bounds.iloc[1, 1], lat[0])
        )
        self.assertTrue(
            np.allclose(res.bounds.iloc[0, 0], -11.800084333105298) or
            np.allclose(res.bounds.iloc[1, 0], -11.800084333105298)
        )
        self.assertTrue(
            np.allclose(res.bounds.iloc[0, 2], np.max(lon)) or
            np.allclose(res.bounds.iloc[1, 2], np.max(lon))
        )
        self.assertTrue(
            np.allclose(res.bounds.iloc[0, 3], np.max(lat)) or
            np.allclose(res.bounds.iloc[1, 3], np.max(lat))
        )

    def test_get_country_geometries_all_pass(self):
        """get_country_geometries with no countries or extent; i.e. the whole
        earth"""
        res = u_coord.get_country_geometries(resolution=110)
        self.assertIsInstance(res, gpd.geodataframe.GeoDataFrame)
        self.assertAlmostEqual(res.area[0], 1.639510995900778)

    def test_country_code_pass(self):
        """Test set_region_id"""

        lon = np.array([-59.6250000000000, -59.6250000000000, -59.6250000000000,
                        -59.5416666666667, -59.5416666666667, -59.4583333333333,
                        -60.2083333333333, -60.2083333333333])
        lat = np.array([13.125, 13.20833333, 13.29166667, 13.125, 13.20833333,
                        13.125, 12.625, 12.70833333])
        for gridded in [True, False]:
            region_id = u_coord.get_country_code(lat, lon, gridded=gridded)
            region_id_OSLO = u_coord.get_country_code(59.91, 10.75, gridded=gridded)
            self.assertEqual(np.count_nonzero(region_id), 6)
            # 052 for barbados
            self.assertTrue(np.all(region_id[:6] == 52))
            # 578 for Norway
            self.assertEqual(region_id_OSLO, np.array([578]))

    def test_get_admin1_info_pass(self):
        """test get_admin1_info()"""
        country_names = ['CHE', 'Indonesia', '840', 51]
        admin1_info, admin1_shapes = u_coord.get_admin1_info(country_names=country_names)
        self.assertEqual(len(admin1_info), 4)
        self.assertListEqual(list(admin1_info.keys()), ['CHE', 'IDN', 'USA', 'ARM'])
        self.assertEqual(len(admin1_info['CHE']), len(admin1_shapes['CHE']))
        self.assertEqual(len(admin1_info['CHE']), 26)
        self.assertEqual(len(admin1_shapes['IDN']), 33)
        self.assertEqual(len(admin1_info['USA']), 51)
        # depending on the version of Natural Earth, this is Washington or Idaho:
        self.assertIn(admin1_info['USA'][1]['iso_3166_2'], ['US-WA', 'US-ID'])

    def test_get_admin1_geometries_pass(self):
        """test get_admin1_geometries"""
        countries = ['CHE', 'Indonesia', '840', 51]
        gdf = u_coord.get_admin1_geometries(countries=countries)
        self.assertEqual(len(gdf.iso_3a.unique()), 4) # 4 countries
        self.assertEqual(gdf.loc[gdf.iso_3a=='CHE'].shape[0], 26) # 26 cantons in CHE
        self.assertEqual(gdf.shape[0], 121) # 121 admin 1 regions in the 4 countries
        self.assertIn('ARM', gdf.iso_3a.values) # Armenia (region_id 051)
        self.assertIn('756', gdf.iso_3n.values) # Switzerland (region_id 756)
        self.assertIn('CH-AI', gdf.iso_3166_2.values) # canton in CHE
        self.assertIn('Sulawesi Tengah', gdf.admin1_name.values) # region in Indonesia
        self.assertIsInstance(gdf.loc[gdf.iso_3166_2 == 'CH-AI'].geometry.values[0],
                              shapely.geometry.MultiPolygon)
        self.assertIsInstance(gdf.loc[gdf.admin1_name == 'Sulawesi Tengah'].geometry.values[0],
                              shapely.geometry.MultiPolygon)
        self.assertIsInstance(gdf.loc[gdf.admin1_name == 'Valais'].geometry.values[0],
                              shapely.geometry.Polygon)

    def test_get_admin1_geometries_fail(self):
        """test get_admin1_geometries wrong input"""
        # non existing country:
        self.assertRaises(LookupError, u_coord.get_admin1_geometries, ["FantasyLand"])
        # wrong variable type for 'countries', e.g. Polygon:
        self.assertRaises(TypeError, u_coord.get_admin1_geometries, shapely.geometry.Polygon())

class TestRasterMeta(unittest.TestCase):
    def test_is_regular_pass(self):
        """Test is_regular function."""
        coord = np.array([[1, 2], [4.4, 5.4], [4, 5]])
        reg, hei, wid = u_coord.grid_is_regular(coord)
        self.assertFalse(reg)
        self.assertEqual(hei, 1)
        self.assertEqual(wid, 1)

        coord = np.array([[1, 2], [4.4, 5], [4, 5]])
        reg, hei, wid = u_coord.grid_is_regular(coord)
        self.assertFalse(reg)
        self.assertEqual(hei, 1)
        self.assertEqual(wid, 1)

        coord = np.array([[1, 2], [4, 5]])
        reg, hei, wid = u_coord.grid_is_regular(coord)
        self.assertFalse(reg)
        self.assertEqual(hei, 1)
        self.assertEqual(wid, 1)

        coord = np.array([[1, 2], [4, 5], [1, 5], [4, 3]])
        reg, hei, wid = u_coord.grid_is_regular(coord)
        self.assertFalse(reg)
        self.assertEqual(hei, 2)
        self.assertEqual(wid, 1)

        coord = np.array([[1, 2], [4, 5], [1, 5], [4, 2]])
        reg, hei, wid = u_coord.grid_is_regular(coord)
        self.assertTrue(reg)
        self.assertEqual(hei, 2)
        self.assertEqual(wid, 2)

        grid_x, grid_y = np.mgrid[10: 100: complex(0, 5),
                                  0: 10: complex(0, 5)]
        grid_x = grid_x.reshape(-1,)
        grid_y = grid_y.reshape(-1,)
        coord = np.array([grid_x, grid_y]).transpose()
        reg, hei, wid = u_coord.grid_is_regular(coord)
        self.assertTrue(reg)
        self.assertEqual(hei, 5)
        self.assertEqual(wid, 5)

        grid_x, grid_y = np.mgrid[10: 100: complex(0, 4),
                                  0: 10: complex(0, 5)]
        grid_x = grid_x.reshape(-1,)
        grid_y = grid_y.reshape(-1,)
        coord = np.array([grid_x, grid_y]).transpose()
        reg, hei, wid = u_coord.grid_is_regular(coord)
        self.assertTrue(reg)
        self.assertEqual(hei, 5)
        self.assertEqual(wid, 4)

    def test_get_resolution_pass(self):
        """Test _get_resolution method"""
        lat = np.array([13.125, 13.20833333, 13.29166667, 13.125,
                        13.20833333, 13.125, 12.625, 12.70833333,
                        12.79166667, 12.875, 12.95833333, 13.04166667])
        lon = np.array([
            -59.6250000000000, -59.6250000000000, -59.6250000000000, -59.5416666666667,
            -59.5416666666667, -59.4583333333333, -60.2083333333333, -60.2083333333333,
            -60.2083333333333, -60.2083333333333, -60.2083333333333, -60.2083333333333
        ])
        res_lat, res_lon = u_coord.get_resolution(lat, lon)
        self.assertAlmostEqual(res_lat, 0.0833333333333)
        self.assertAlmostEqual(res_lon, 0.0833333333333)

    def test_vector_to_raster_pass(self):
        """Test vector_to_raster"""
        xmin, ymin, xmax, ymax = -60, -5, -50, 10  # bounds of points == centers pixels
        points_bounds = (xmin, ymin, xmax, ymax)
        res = 0.5
        rows, cols, ras_trans = u_coord.pts_to_raster_meta(points_bounds, (res, -res))
        self.assertEqual(xmin - res / 2 + res * cols, xmax + res / 2)
        self.assertEqual(ymax + res / 2 - res * rows, ymin - res / 2)
        self.assertEqual(ras_trans[0], res)
        self.assertEqual(ras_trans[4], -res)
        self.assertEqual(ras_trans[1], 0.0)
        self.assertEqual(ras_trans[3], 0.0)
        self.assertEqual(ras_trans[2], xmin - res / 2)
        self.assertEqual(ras_trans[5], ymax + res / 2)
        self.assertTrue(ymin >= ymax + res / 2 - rows * res)
        self.assertTrue(xmax <= xmin - res / 2 + cols * res)

    def test_pts_to_raster_irreg_pass(self):
        """Test pts_to_raster_meta with irregular points"""
        # bounds of points == centers of pixels
        points_bounds = (-124.19473, 32.81908, -114.4632, 42.020759999999996)
        xmin, ymin, xmax, ymax = points_bounds
        res = 0.013498920086393088
        rows, cols, ras_trans = u_coord.pts_to_raster_meta(points_bounds, (res, -res))
        self.assertEqual(ras_trans[0], res)
        self.assertEqual(ras_trans[4], -res)
        self.assertEqual(ras_trans[1], 0.0)
        self.assertEqual(ras_trans[3], 0.0)
        self.assertEqual(ras_trans[2], xmin - res / 2)
        self.assertEqual(ras_trans[5], ymax + res / 2)
        self.assertTrue(ymin >= ymax + res / 2 - rows * res)
        self.assertTrue(xmax <= xmin - res / 2 + cols * res)

    def test_points_to_raster_pass(self):
        """Test points_to_raster"""
        df_val = gpd.GeoDataFrame()
        x, y = np.meshgrid(np.linspace(0, 2, 5), np.linspace(40, 50, 10))
        df_val['latitude'] = y.flatten()
        df_val['longitude'] = x.flatten()
        df_val['value'] = np.ones(len(df_val)) * 10
        crs = 'epsg:2202'
        _raster, meta = u_coord.points_to_raster(df_val, val_names=['value'], crs=crs)
        self.assertIsNone(df_val.crs)  # points_to_raster must not modify df_val
        self.assertTrue(u_coord.equal_crs(meta['crs'], crs))
        self.assertAlmostEqual(meta['transform'][0], 0.5)
        self.assertAlmostEqual(meta['transform'][1], 0)
        self.assertAlmostEqual(meta['transform'][2], -0.25)
        self.assertAlmostEqual(meta['transform'][3], 0)
        self.assertAlmostEqual(meta['transform'][4], -0.5)
        self.assertAlmostEqual(meta['transform'][5], 50.25)
        self.assertEqual(meta['height'], 21)
        self.assertEqual(meta['width'], 5)

        # test for values crossing antimeridian
        df_val = gpd.GeoDataFrame()
        df_val['latitude'] = [1, 0, 1, 0]
        df_val['longitude'] = [178, -179.0, 181, -180]
        df_val['value'] = np.arange(4)
        r_data, meta = u_coord.points_to_raster(
            df_val, val_names=['value'], res=0.5, raster_res=1.0)
        self.assertTrue(u_coord.equal_crs(meta['crs'], DEF_CRS))
        self.assertAlmostEqual(meta['transform'][0], 1.0)
        self.assertAlmostEqual(meta['transform'][1], 0)
        self.assertAlmostEqual(meta['transform'][2], 177.5)
        self.assertAlmostEqual(meta['transform'][3], 0)
        self.assertAlmostEqual(meta['transform'][4], -1.0)
        self.assertAlmostEqual(meta['transform'][5], 1.5)
        self.assertEqual(meta['height'], 2)
        self.assertEqual(meta['width'], 4)
        np.testing.assert_array_equal(r_data[0], [[0, 0, 0, 2], [0, 0, 3, 1]])

class TestRasterIO(unittest.TestCase):
    def test_write_raster_pass(self):
        """Test write_raster function."""
        test_file = Path(DATA_DIR, "test_write_raster.tif")
        data = np.arange(24).reshape(6, 4).astype(np.float32)
        meta = {
            'transform': Affine(0.1, 0, 0, 0, 1, 0),
            'width': data.shape[1],
            'height': data.shape[0],
            'crs': 'epsg:2202',
            'compress': 'deflate',
        }
        u_coord.write_raster(test_file, data, meta)
        read_meta, read_data = u_coord.read_raster(test_file)
        self.assertEqual(read_meta['transform'], meta['transform'])
        self.assertEqual(read_meta['width'], meta['width'])
        self.assertEqual(read_meta['height'], meta['height'])
        self.assertTrue(u_coord.equal_crs(read_meta['crs'], meta['crs']))
        self.assertEqual(read_data.shape, (1, np.prod(data.shape)))
        np.testing.assert_array_equal(read_data, data.reshape(read_data.shape))

    def test_window_raster_pass(self):
        """Test window"""
        meta, inten_ras = u_coord.read_raster(HAZ_DEMO_FL, window=Window(10, 20, 50.1, 60))
        self.assertAlmostEqual(meta['crs'], DEF_CRS)
        self.assertAlmostEqual(meta['transform'].c, -69.2471495969998)
        self.assertAlmostEqual(meta['transform'].a, 0.009000000000000341)
        self.assertAlmostEqual(meta['transform'].b, 0.0)
        self.assertAlmostEqual(meta['transform'].f, 10.248220966978932)
        self.assertAlmostEqual(meta['transform'].d, 0.0)
        self.assertAlmostEqual(meta['transform'].e, -0.009000000000000341)
        self.assertEqual(meta['height'], 60)
        self.assertEqual(meta['width'], 50)
        self.assertEqual(inten_ras.shape, (1, 60 * 50))
        self.assertAlmostEqual(inten_ras.reshape((60, 50))[25, 12], 0.056825936)

    def test_poly_raster_pass(self):
        """Test geometry"""
        poly = box(-69.2471495969998, 9.708220966978912, -68.79714959699979, 10.248220966978932)
        meta, inten_ras = u_coord.read_raster(HAZ_DEMO_FL, geometry=[poly])
        self.assertAlmostEqual(meta['crs'], DEF_CRS)
        self.assertAlmostEqual(meta['transform'].c, -69.2471495969998)
        self.assertAlmostEqual(meta['transform'].a, 0.009000000000000341)
        self.assertAlmostEqual(meta['transform'].b, 0.0)
        self.assertAlmostEqual(meta['transform'].f, 10.248220966978932)
        self.assertAlmostEqual(meta['transform'].d, 0.0)
        self.assertAlmostEqual(meta['transform'].e, -0.009000000000000341)
        self.assertEqual(meta['height'], 60)
        self.assertEqual(meta['width'], 50)
        self.assertEqual(inten_ras.shape, (1, 60 * 50))

    def test_crs_raster_pass(self):
        """Test change projection"""
        meta, inten_ras = u_coord.read_raster(
            HAZ_DEMO_FL, dst_crs='epsg:2202', resampling=Resampling.nearest)
        self.assertAlmostEqual(meta['crs'], 'epsg:2202')
        self.assertAlmostEqual(meta['transform'].c, 462486.8490210658)
        self.assertAlmostEqual(meta['transform'].a, 998.576177833903)
        self.assertAlmostEqual(meta['transform'].b, 0.0)
        self.assertAlmostEqual(meta['transform'].f, 1164831.4772731226)
        self.assertAlmostEqual(meta['transform'].d, 0.0)
        self.assertAlmostEqual(meta['transform'].e, -998.576177833903)
        self.assertEqual(meta['height'], 1081)
        self.assertEqual(meta['width'], 968)
        self.assertEqual(inten_ras.shape, (1, 1081 * 968))
        # TODO: NOT RESAMPLING WELL in this case!?
        self.assertAlmostEqual(inten_ras.reshape((1081, 968))[45, 22], 0)

    def test_crs_and_geometry_raster_pass(self):
        """Test change projection and crop to geometry"""
        ply = shapely.geometry.Polygon([
            (478080.8562247154, 1105419.13439131),
            (478087.5912452241, 1116475.583523723),
            (500000, 1116468.876713805),
            (500000, 1105412.49126517),
            (478080.8562247154, 1105419.13439131)
        ])
        meta, inten_ras = u_coord.read_raster(
            HAZ_DEMO_FL, dst_crs='epsg:2202', geometry=[ply],
            resampling=Resampling.nearest)
        self.assertAlmostEqual(meta['crs'], 'epsg:2202')
        self.assertEqual(meta['height'], 12)
        self.assertEqual(meta['width'], 23)
        self.assertEqual(inten_ras.shape, (1, 12 * 23))
        # TODO: NOT RESAMPLING WELL in this case!?
        self.assertAlmostEqual(inten_ras.reshape((12, 23))[11, 12], 0.10063865780830383)

    def test_transform_raster_pass(self):
        transform = Affine(0.009000000000000341, 0.0, -69.33714959699981,
                           0.0, -0.009000000000000341, 10.42822096697894)
        meta, inten_ras = u_coord.read_raster(
            HAZ_DEMO_FL, transform=transform, height=500, width=501)

        left = meta['transform'].xoff
        top = meta['transform'].yoff
        bottom = top + meta['transform'][4] * meta['height']
        right = left + meta['transform'][0] * meta['width']

        self.assertAlmostEqual(left, -69.33714959699981)
        self.assertAlmostEqual(bottom, 5.928220966978939)
        self.assertAlmostEqual(right, -64.82814959699981)
        self.assertAlmostEqual(top, 10.42822096697894)
        self.assertEqual(meta['width'], 501)
        self.assertEqual(meta['height'], 500)
        self.assertTrue(u_coord.equal_crs(meta['crs'].to_epsg(), 4326))
        self.assertEqual(inten_ras.shape, (1, 500 * 501))

        meta, inten_all = u_coord.read_raster(HAZ_DEMO_FL, window=Window(0, 0, 501, 500))
        self.assertTrue(np.array_equal(inten_all, inten_ras))

    def test_sample_raster(self):
        """Test sampling points from raster file"""
        val_1, val_2, fill_value = 0.056825936, 0.10389626, -999
        i_j_vals = np.array([
            [44, 21, 0],
            [44, 22, 0],
            [44, 23, 0],
            [45, 21, 0],
            [45, 22, val_1],
            [45, 23, val_2],
            [46, 21, 0],
            [46, 22, 0],
            [46, 23, 0],
            [45, 22.2, 0.8 * val_1 + 0.2 * val_2],
            [45.3, 21.4, 0.7 * 0.4 * val_1],
            [-20, 0, fill_value],
        ])
        res = 0.009000000000000341
        lat = 10.42822096697894 - res / 2 - i_j_vals[:, 0] * res
        lon = -69.33714959699981 + res / 2 + i_j_vals[:, 1] * res
        values = u_coord.read_raster_sample(HAZ_DEMO_FL, lat, lon, fill_value=fill_value)
        self.assertEqual(values.size, lat.size)
        for i, val in enumerate(i_j_vals[:, 2]):
            self.assertAlmostEqual(values[i], val)

        # with explicit intermediate resolution
        values = u_coord.read_raster_sample(HAZ_DEMO_FL, lat, lon, fill_value=fill_value,
                                    intermediate_res=res)
        self.assertEqual(values.size, lat.size)
        for i, val in enumerate(i_j_vals[:, 2]):
            self.assertAlmostEqual(values[i], val)

    def test_refine_raster(self):
        """Test refinement of given raster data"""
        data = np.array([
            [0.25, 0.75],
            [0.5, 1],
        ])
        transform = Affine(0.5, 0, 0, 0, 0.5, 0)
        new_res = 0.1
        new_data, new_transform = u_coord.refine_raster_data(data, transform, new_res)

        self.assertEqual(new_transform[0], new_res)
        self.assertEqual(new_transform[4], new_res)
        self.assertAlmostEqual(new_data[2, 2], data[0, 0])
        self.assertAlmostEqual(new_data[2, 7], data[0, 1])
        self.assertAlmostEqual(new_data[7, 2], data[1, 0])
        self.assertAlmostEqual(new_data[7, 7], data[1, 1])
        self.assertAlmostEqual(new_data[1, 2], data[0, 0])
        self.assertAlmostEqual(new_data[3, 3], 0.4)

    def test_bounded_refined_raster(self):
        """Test reading a raster within specified bounds and at specified resolution"""
        # the bounds have an additional digit (1) at the end to avoid floating-point errors
        bounds = (-69.141, 9.991, -69.111, 10.021)
        res = 0.004
        global_origin = (-180, 90)
        z, transform = u_coord.read_raster_bounds(
            HAZ_DEMO_FL, bounds, res=res, global_origin=global_origin)

        # the first dimension corresponds to the raster bands:
        self.assertEqual(z.shape[0], 1)
        z = z[0]

        # the signs of stepsizes are retained from the original raster:
        self.assertLess(transform[4], 0)
        self.assertGreater(transform[0], 0)

        # the absolute step sizes are as requested
        self.assertEqual(np.abs(transform[0]), res)
        self.assertEqual(np.abs(transform[4]), res)

        # the raster is aligned to the specified global origin
        align_x = (transform[2] - global_origin[0]) / res
        align_y = (transform[5] - global_origin[1]) / res
        self.assertAlmostEqual(align_x, np.round(align_x))
        self.assertAlmostEqual(align_y, np.round(align_y))

        # the bounds of returned pixel centers cover the specified region:
        # check along x-axis
        self.assertLessEqual(transform[2] + 0.5 * transform[0], bounds[0])
        self.assertGreater(transform[2] + 1.5 * transform[0], bounds[0])
        self.assertGreaterEqual(transform[2] + (z.shape[1] - 0.5) * transform[0], bounds[2])
        self.assertLess(transform[2] + (z.shape[1] - 1.5) * transform[0], bounds[2])

        # check along y-axis (note that the orientation is reversed)
        self.assertGreaterEqual(transform[5] + 0.5 * transform[4], bounds[3])
        self.assertLess(transform[5] + 1.5 * transform[4], bounds[3])
        self.assertLessEqual(transform[5] + (z.shape[0] - 0.5) * transform[4], bounds[1])
        self.assertGreater(transform[5] + (z.shape[0] - 1.5) * transform[4], bounds[1])

        # trigger downloading of dist-to-coast dataset (if not already present)
        path = u_coord._get_dist_to_coast_nasa_tif()

        # make sure the buffering doesn't go beyond ±90 degrees latitude:
        z, transform = u_coord.read_raster_bounds(
            path, (0, -90, 10, -80), res=1.0, global_origin=(-180, 90))
        self.assertEqual(z.shape, (1, 11, 12))
        self.assertEqual(transform[5], -79.0)
        z, transform = u_coord.read_raster_bounds(
            path, (0, 80, 10, 90), res=1.0, global_origin=(-180, 90))
        self.assertEqual(z.shape, (1, 11, 12))
        self.assertEqual(transform[5], 90.0)

        # make sure crossing the antimeridian works fine:
        z_right, transform = u_coord.read_raster_bounds(
            path, (-175, 0, -170, 10), res=1.0, global_origin=(-180, 90))
        z_left, transform = u_coord.read_raster_bounds(
            path, (170, 0, 175, 10), res=1.0, global_origin=(-180, 90))
        z_both, transform = u_coord.read_raster_bounds(
            path, (170, 0, 190, 10), res=1.0, global_origin=(-180, 90))
        z_both_neg, transform = u_coord.read_raster_bounds(
            path, (-190, 0, -170, 10), res=1.0, global_origin=(-180, 90))
        np.testing.assert_array_equal(z_left[0,:,:], z_both[0,:,:z_left.shape[2]])
        np.testing.assert_array_equal(z_right[0,:,:], z_both[0,:,-z_right.shape[2]:])
        np.testing.assert_array_equal(z_left[0,:,:], z_both_neg[0,:,:z_left.shape[2]])
        np.testing.assert_array_equal(z_right[0,:,:], z_both_neg[0,:,-z_right.shape[2]:])

    def test_subraster_from_bounds(self):
        """test subraster_from_bounds function"""
        transform = rasterio.transform.from_origin(-10, 10, 0.5, 0.5)
        bounds = (-3.4, 1.2, 0.6, 4.26)
        dst_transform, dst_shape = u_coord.subraster_from_bounds(transform, bounds)
        self.assertEqual(dst_shape, (7, 8))
        self.assertEqual(dst_transform[:6], (0.5, 0, -3.5, 0, -0.5, 4.5))

        # test whether orientation is preserved
        transform = rasterio.transform.from_origin(-10, 10, 0.5, -0.5)
        dst_transform, dst_shape = u_coord.subraster_from_bounds(transform, bounds)
        self.assertEqual(dst_shape, (7, 8))
        self.assertEqual(dst_transform[:6], (0.5, 0, -3.5, 0, 0.5, 1.0))

        # test for more complicated input data:
        _, meta_list = data_arrays_resampling_demo()
        i = 2
        dst_resolution = (1., .2)
        bounds = rasterio.transform.array_bounds(
            meta_list[i]['height'], meta_list[i]['width'], meta_list[i]['transform'])
        transform = rasterio.transform.from_origin(-180, 90, *dst_resolution)
        dst_transform, dst_shape = u_coord.subraster_from_bounds(transform, bounds)
        self.assertEqual(dst_shape, (meta_list[0]['height'] / dst_resolution[1],
                                     meta_list[0]['width'] / dst_resolution[0]))
        self.assertEqual(dst_resolution, (dst_transform[0], -dst_transform[4]))
        self.assertEqual(meta_list[i]['transform'][1], dst_transform[1])
        self.assertEqual(meta_list[i]['transform'][2], dst_transform[2])
        self.assertEqual(meta_list[i]['transform'][3], dst_transform[3])
        self.assertEqual(meta_list[i]['transform'][5], dst_transform[5])

        # test for odd resolution change:
        i = 0
        dst_resolution = (.15, .15)
        bounds = rasterio.transform.array_bounds(
            meta_list[i]['height'], meta_list[i]['width'], meta_list[i]['transform'])
        transform = rasterio.transform.from_origin(-180, 90, *dst_resolution)
        dst_transform, dst_shape = u_coord.subraster_from_bounds(transform, bounds)
        self.assertEqual(dst_shape, (14, 20))
        self.assertEqual(transform[0], dst_transform[0])
        self.assertEqual(transform[4], dst_transform[4])
        self.assertAlmostEqual(-10.05, dst_transform[2])
        self.assertAlmostEqual(40.05, dst_transform[5])

    def test_align_raster_data_shift(self):
        """test function align_raster_data for geographical shift"""
        data_in, meta_list = data_arrays_resampling_demo()
        i = 0  # dst
        j = 1  # src

        dst_resolution=meta_list[i]['transform'][0]
        dst_bounds = rasterio.transform.array_bounds(
            meta_list[i]['height'], meta_list[i]['width'], meta_list[i]['transform'])
        data_out, dst_transform = u_coord.align_raster_data(
            data_in[j], meta_list[j]['crs'], meta_list[j]['transform'],
            dst_crs=meta_list[i]['crs'], dst_resolution=dst_resolution, dst_bounds=dst_bounds,
            resampling='bilinear')

        # test northward shift of box:
        np.testing.assert_array_equal(data_in[1][1,:], data_out[0,:])
        np.testing.assert_array_equal(np.array([0., 0., 0.], dtype='float32'), data_out[1,:])
        self.assertEqual(meta_list[i]['transform'][5], dst_transform[5])

    def test_align_raster_data_downsampling(self):
        """test function align_raster_data for downsampling"""
        data_in, meta_list = data_arrays_resampling_demo()
        i = 0  # dst
        j = 2  # src

        dst_resolution=meta_list[i]['transform'][0]
        dst_bounds = rasterio.transform.array_bounds(
            meta_list[i]['height'], meta_list[i]['width'], meta_list[i]['transform'])
        data_out, dst_transform = u_coord.align_raster_data(
            data_in[j], meta_list[j]['crs'], meta_list[j]['transform'],
            dst_crs=meta_list[i]['crs'], dst_resolution=dst_resolution, dst_bounds=dst_bounds,
            resampling='bilinear')

        # test downsampled data:
        reference_array = np.array([[5.0204080, 2.2678570, 0.12244898],
                                    [1.1224489, 0.6785714, 0.73469390]], dtype='float32')
        np.testing.assert_array_almost_equal_nulp(reference_array, data_out)
        self.assertEqual(dst_resolution, dst_transform[0])

    def test_align_raster_data_downsample_conserve(self):
        """test function align_raster_data downsampling with conservation
        of mean and sum and normalization"""
        data_in, meta_list = data_arrays_resampling_demo()
        i = 0  # dst

        dst_resolution=meta_list[i]['transform'][0]
        dst_bounds = rasterio.transform.array_bounds(
            meta_list[i]['height'], meta_list[i]['width'], meta_list[i]['transform'])

        # test conserve sum:
        for j, data in enumerate(data_in): # src
            data_out, _ = u_coord.align_raster_data(
                data_in[j], meta_list[j]['crs'], meta_list[j]['transform'],
                dst_crs=meta_list[i]['crs'], dst_resolution=dst_resolution, dst_bounds=dst_bounds,
                resampling='bilinear', conserve='sum')
            self.assertAlmostEqual(data_in[j].sum(), data_out.sum(), places=4)

        # test conserve mean:
        for j, data in enumerate(data_in):
            data_out, _ = u_coord.align_raster_data(
                data_in[j], meta_list[j]['crs'], meta_list[j]['transform'],
                dst_crs=meta_list[i]['crs'], dst_resolution=dst_resolution, dst_bounds=dst_bounds,
                resampling='bilinear', conserve='mean')
            self.assertAlmostEqual(data_in[j].mean(), data_out.mean(), places=4)

    def test_align_raster_data_upsample(self):
        """test function align_raster_data with upsampling"""
        data_in, meta_list = data_arrays_resampling_demo()
        data_out = list()
        i = 2  # dst

        dst_resolution = meta_list[i]['transform'][0]
        dst_bounds = rasterio.transform.array_bounds(
            meta_list[i]['height'], meta_list[i]['width'], meta_list[i]['transform'])
        for j in [0,1,2]:
            data_out.append(u_coord.align_raster_data(
                data_in[j], meta_list[j]['crs'], meta_list[j]['transform'],
                dst_crs=meta_list[i]['crs'], dst_resolution=dst_resolution,
                dst_bounds=dst_bounds, resampling='bilinear')[0])

        # test reference data unchanged:
        np.testing.assert_array_equal(data_in[2], data_out[2])
        # test northward shift:
        np.testing.assert_array_equal(data_out[0][2,:], data_out[1][0,:])
        np.testing.assert_array_equal(data_out[0][3,:], data_out[1][1,:])
        # test upsampled data:
        reference_array = np.array([[0.00, 0.25, 0.75, 1.25, 1.75, 2.00],
                                    [0.75, 1.00, 1.50, 2.00, 2.50, 2.75],
                                    [2.25, 2.50, 3.00, 3.50, 4.00, 4.25],
                                    [3.00, 3.25, 3.75, 4.25, 4.75, 5.00]], dtype='float32')
        np.testing.assert_array_equal(reference_array, data_out[0])

    def test_align_raster_data_odd_downsample(self):
        """test function resample_input_data with odd downsampling"""
        data_in, meta_list = data_arrays_resampling_demo()
        i = 0
        j = 0

        dst_resolution = 1.7
        dst_bounds = rasterio.transform.array_bounds(
            meta_list[i]['height'], meta_list[i]['width'], meta_list[i]['transform'])
        data_out, dst_transform = u_coord.align_raster_data(
            data_in[j], meta_list[j]['crs'], meta_list[j]['transform'],
            dst_crs=meta_list[i]['crs'], dst_resolution=dst_resolution, dst_bounds=dst_bounds,
            resampling='bilinear')

        self.assertEqual(dst_resolution, dst_transform[0])
        reference_array = np.array([[0.425, 1.7631578],
                                    [3.425, 4.763158 ]], dtype='float32')
        np.testing.assert_array_equal(reference_array, data_out)

    def test_mask_raster_with_geometry(self):
        """test function mask_raster_with_geometry"""
        raster = np.ones((4, 3), dtype=np.float32)
        transform = rasterio.transform.Affine(1, 0, 5, 0, -1, -10)
        shapes = [shapely.geometry.box(6.1, -12.9, 6.9, -11.1)]
        expected = np.array([
                            [0, 0, 0],
                            [0, 1, 0],
                            [0, 1, 0],
                            [0, 0, 0],
                            ], dtype=np.float32)
        np.testing.assert_array_equal(
            u_coord.mask_raster_with_geometry(raster, transform, shapes), expected)


# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestFunc)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestAssign))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestGetGeodata))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestRasterMeta))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestRasterIO))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define Measure class.
"""

__all__ = ['Measure']

import copy
import logging
from pathlib import Path
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame

from climada.entity.exposures.base import Exposures, INDICATOR_IMPF, INDICATOR_CENTR
import climada.util.checker as u_check

LOGGER = logging.getLogger(__name__)

IMPF_ID_FACT = 1000
"""Factor internally used as id for impact functions when region selected."""

NULL_STR = 'nil'
"""String considered as no path in measures exposures_set and hazard_set or
no string in imp_fun_map"""

class Measure():
    """
    Contains the definition of one measure.

    Attributes
    ----------
    name : str
        name of the measure
    haz_type : str
        related hazard type (peril), e.g. TC
    color_rgb : np.array
        integer array of size 3. Color code of this measure in RGB
    cost : float
        discounted cost (in same units as assets)
    hazard_set : str
        file name of hazard to use (in h5 format)
    hazard_freq_cutoff : float
        hazard frequency cutoff
    exposures_set : str  or climada.entity.Exposure
        file name of exposure to use (in h5 format) or Exposure instance
    imp_fun_map : str
        change of impact function id of exposures, e.g. '1to3'
    hazard_inten_imp : tuple(float, float)
        parameter a and b of hazard intensity change
    mdd_impact : tuple(float, float)
        parameter a and b of the impact over the mean damage degree
    paa_impact : tuple(float, float)
        parameter a and b of the impact over the percentage of affected assets
    exp_region_id : int
        region id of the selected exposures to consider ALL the previous
        parameters
    risk_transf_attach : float
        risk transfer attachment
    risk_transf_cover : float
        risk transfer cover
    risk_transf_cost_factor : float
        factor to multiply to resulting insurance layer to get the total
        cost of risk transfer
    """

    def __init__(self):
        """Empty initialization."""
        self.name = ''
        self.haz_type = ''
        self.color_rgb = np.array([0, 0, 0])
        self.cost = 0

        # related to change in hazard
        self.hazard_set = NULL_STR
        self.hazard_freq_cutoff = 0

        # related to change in exposures
        self.exposures_set = NULL_STR
        self.imp_fun_map = NULL_STR  # ids of impact functions to change e.g. 1to10

        # related to change in impact functions
        self.hazard_inten_imp = (1, 0)  # parameter a and b
        self.mdd_impact = (1, 0)  # parameter a and b
        self.paa_impact = (1, 0)  # parameter a and b

        # related to change in region
        self.exp_region_id = []

        # risk transfer
        self.risk_transf_attach = 0
        self.risk_transf_cover = 0
        self.risk_transf_cost_factor = 1

    def check(self):
        """
        Check consistent instance data.

        Raises
        ------
        ValueError
        """
        u_check.size([3, 4], self.color_rgb, 'Measure.color_rgb')
        u_check.size(2, self.hazard_inten_imp, 'Measure.hazard_inten_imp')
        u_check.size(2, self.mdd_impact, 'Measure.mdd_impact')
        u_check.size(2, self.paa_impact, 'Measure.paa_impact')

    def calc_impact(self, exposures, imp_fun_set, hazard):
        """
        Apply measure and compute impact and risk transfer of measure
        implemented over inputs.

        Parameters
        ----------
        exposures : climada.entity.Exposures
            exposures instance
        imp_fun_set : climada.entity.ImpactFuncSet
            impact function set instance
        hazard : climada.hazard.Hazard
            hazard instance

        Returns
        -------
            : climada.engine.Impact
            resulting impact and risk transfer of measure
        """

        new_exp, new_impfs, new_haz = self.apply(exposures, imp_fun_set, hazard)
        return self._calc_impact(new_exp, new_impfs, new_haz)

    def apply(self, exposures, imp_fun_set, hazard):
        """
        Implement measure with all its defined parameters.

        Parameters
        ----------
        exposures : climada.entity.Exposures
            exposures instance
        imp_fun_set : climada.entity.ImpactFuncSet
            impact function set instance
        hazard : climada.hazard.Hazard
            hazard instance

        Returns
        -------
        new_exp, new_ifs, new_haz : climada.entity.Exposure,
                                    climada.entity.ImpactFuncSet,
                                    climada.hazard.Hazard
            Exposure, impact function set with implemented measure
            with all defined parameters.
        """
        # change hazard
        new_haz = self._change_all_hazard(hazard)
        # change exposures
        new_exp = self._change_all_exposures(exposures)
        new_exp = self._change_exposures_impf(new_exp)
        # change impact functions
        new_impfs = self._change_imp_func(imp_fun_set)
        # cutoff events whose damage happen with high frequency (in region impf specified)
        new_haz = self._cutoff_hazard_damage(new_exp, new_impfs, new_haz)
        # apply all previous changes only to the selected exposures
        new_exp, new_impfs, new_haz = self._filter_exposures(
            exposures, imp_fun_set, hazard, new_exp, new_impfs, new_haz)

        return new_exp, new_impfs, new_haz

    def _calc_impact(self, new_exp, new_impfs, new_haz):
        """Compute impact and risk transfer of measure implemented over inputs.

        Parameters
        ----------
        new_exp : climada.entity.Exposures
            exposures once measure applied
        new_ifs : climada.entity.ImpactFuncSet
            impact function set once measure applied
        new_haz  : climada.hazard.Hazard
            hazard once measure applied

        Returns
        -------
            : climada.engine.Impact
        """
        from climada.engine.impact import Impact
        imp = Impact()
        imp.calc(new_exp, new_impfs, new_haz)
        return imp.calc_risk_transfer(self.risk_transf_attach, self.risk_transf_cover)

    def _change_all_hazard(self, hazard):
        """
        Change hazard to provided hazard_set.

        Parameters
        ----------
        hazard : climada.hazard.Hazard
            hazard instance

        Returns
        -------
        new_haz : climada.hazard.Hazard
            Hazard
        """
        if self.hazard_set == NULL_STR:
            return hazard

        LOGGER.debug('Setting new hazard %s', self.hazard_set)
        from climada.hazard.base import Hazard
        new_haz = Hazard.from_hdf5(self.hazard_set)
        new_haz.check()
        return new_haz

    def _change_all_exposures(self, exposures):
        """
        Change exposures to provided exposures_set.

        Parameters
        ----------
        exposures : climada.entity.Exposures
            exposures instance

        Returns
        -------
        new_exp : climada.entity.Exposures()
            Exposures
        """
        if isinstance(self.exposures_set, str) and self.exposures_set == NULL_STR:
            return exposures

        if isinstance(self.exposures_set, (str, Path)):
            LOGGER.debug('Setting new exposures %s', self.exposures_set)
            new_exp = Exposures.from_hdf5(self.exposures_set)
            new_exp.check()
        elif isinstance(self.exposures_set, Exposures):
            LOGGER.debug('Setting new exposures. ')
            new_exp = self.exposures_set.copy(deep=True)
            new_exp.check()
        else:
            raise ValueError(f'{self.exposures_set} is neither a string nor an Exposures object')

        if not np.array_equal(np.unique(exposures.gdf.latitude.values),
                              np.unique(new_exp.gdf.latitude.values)) or \
        not np.array_equal(np.unique(exposures.gdf.longitude.values),
                           np.unique(new_exp.gdf.longitude.values)):
            LOGGER.warning('Exposures locations have changed.')

        return new_exp

    def _change_exposures_impf(self, exposures):
        """Change exposures impact functions ids according to imp_fun_map.

        Parameters
        ----------
        exposures : climada.entity.Exposures
            exposures instance

        Returns
        -------
        new_exp : climada.entity.Exposure
            Exposure with updated impact functions ids accordgin to
            impf_fun_map
        """
        if self.imp_fun_map == NULL_STR:
            return exposures

        LOGGER.debug('Setting new exposures impact functions%s', self.imp_fun_map)
        new_exp = exposures.copy(deep=True)
        from_id = int(self.imp_fun_map[0:self.imp_fun_map.find('to')])
        to_id = int(self.imp_fun_map[self.imp_fun_map.find('to') + 2:])
        try:
            exp_change = np.argwhere(new_exp.gdf[INDICATOR_IMPF + self.haz_type].values == from_id).\
                reshape(-1)
            new_exp.gdf[INDICATOR_IMPF + self.haz_type].values[exp_change] = to_id
        except KeyError:
            exp_change = np.argwhere(new_exp.gdf[INDICATOR_IMPF].values == from_id).\
                reshape(-1)
            new_exp.gdf[INDICATOR_IMPF].values[exp_change] = to_id
        return new_exp

    def _change_imp_func(self, imp_set):
        """
        Apply measure to impact functions of the same hazard type.

        Parameters
        ----------
        imp_set : climada.entity.ImpactFuncSet
            impact function set instance to be modified

        Returns
        -------
        new_imp_set : climada.entity.ImpactFuncSet
            ImpactFuncSet with measure applied to each impact function
            according to the defined hazard type
        """
        if self.hazard_inten_imp == (1, 0) and self.mdd_impact == (1, 0)\
        and self.paa_impact == (1, 0):
            return imp_set

        new_imp_set = copy.deepcopy(imp_set)
        for imp_fun in new_imp_set.get_func(self.haz_type):
            LOGGER.debug('Transforming impact functions.')
            imp_fun.intensity = np.maximum(
                imp_fun.intensity * self.hazard_inten_imp[0] - self.hazard_inten_imp[1], 0.0)
            imp_fun.mdd = np.maximum(
                imp_fun.mdd * self.mdd_impact[0] + self.mdd_impact[1], 0.0)
            imp_fun.paa = np.maximum(
                imp_fun.paa * self.paa_impact[0] + self.paa_impact[1], 0.0)

        if not new_imp_set.size():
            LOGGER.info('No impact function of hazard %s found.', self.haz_type)

        return new_imp_set

    def _cutoff_hazard_damage(self, exposures, impf_set, hazard):
        """Cutoff of hazard events which generate damage with a frequency higher
        than hazard_freq_cutoff.

        Parameters
        ----------
        exposures : climada.entity.Exposures
            exposures instance
        imp_set : climada.entity.ImpactFuncSet
            impact function set instance
        hazard : climada.hazard.Hazard
            hazard instance

        Returns
        -------
        new_haz : climada.hazard.Hazard
            Hazard without events which generate damage with a frequency
            higher than hazard_freq_cutoff
        """
        if self.hazard_freq_cutoff == 0:
            return hazard

        if self.exp_region_id:
            # compute impact only in selected region
            in_reg = np.logical_or.reduce(
                [exposures.gdf.region_id.values == reg for reg in self.exp_region_id]
            )
            exp_imp = Exposures(exposures.gdf[in_reg], crs=exposures.crs)
        else:
            exp_imp = exposures

        from climada.engine.impact import Impact
        imp = Impact()
        imp.calc(exp_imp, impf_set, hazard)

        LOGGER.debug('Cutting events whose damage have a frequency > %s.',
                     self.hazard_freq_cutoff)
        new_haz = copy.deepcopy(hazard)
        sort_idxs = np.argsort(imp.at_event)[::-1]
        exceed_freq = np.cumsum(imp.frequency[sort_idxs])
        cutoff = exceed_freq > self.hazard_freq_cutoff
        sel_haz = sort_idxs[cutoff]
        for row in sel_haz:
            new_haz.intensity.data[new_haz.intensity.indptr[row]:
                                   new_haz.intensity.indptr[row + 1]] = 0
        new_haz.intensity.eliminate_zeros()
        return new_haz

    def _filter_exposures(self, exposures, imp_set, hazard, new_exp, new_impfs,
                          new_haz):
        """
        Incorporate changes of new elements to previous ones only for the
        selected exp_region_id. If exp_region_id is [], all new changes
        will be accepted.

        Parameters
        ----------
        exposures : climada.entity.Exposures
            old exposures instance
        imp_set :climada.entity.ImpactFuncSet
            old impact function set instance
        hazard : climada.hazard.Hazard
            old hazard instance
        new_exp : climada.entity.Exposures
            new exposures instance
        new_ifs : climada.entity.ImpactFuncSet
            new impact functions instance
        new_haz : climada.hazard.Hazard
            new hazard instance

        Returns
        -------
        new_exp,new_ifs, new_haz : climada.entity.Exposures,
                                   climada.entity.ImpactFuncSet,
                                   climada.hazard.Hazard
            Exposures, ImpactFuncSet, Hazard with incoporated elements
            for the selected exp_region_id.
        """
        if not self.exp_region_id:
            return new_exp, new_impfs, new_haz

        if exposures is new_exp:
            new_exp = exposures.copy(deep=True)

        if imp_set is not new_impfs:
            # provide new impact functions ids to changed impact functions
            fun_ids = list(new_impfs.get_func()[self.haz_type].keys())
            for key in fun_ids:
                new_impfs.get_func()[self.haz_type][key].id = key + IMPF_ID_FACT
                new_impfs.get_func()[self.haz_type][key + IMPF_ID_FACT] = \
                    new_impfs.get_func()[self.haz_type][key]
            try:
                new_exp.gdf[INDICATOR_IMPF + self.haz_type] += IMPF_ID_FACT
            except KeyError:
                new_exp.gdf[INDICATOR_IMPF] += IMPF_ID_FACT
            # collect old impact functions as well (used by exposures)
            new_impfs.get_func()[self.haz_type].update(imp_set.get_func()[self.haz_type])

        # get the indices for changing and inert regions
        chg_reg = exposures.gdf.region_id.isin(self.exp_region_id)
        no_chg_reg = ~chg_reg

        LOGGER.debug('Number of changed exposures: %s', chg_reg.sum())

        # concatenate previous and new exposures
        new_exp.set_gdf(
            GeoDataFrame(
                pd.concat([
                    exposures.gdf[no_chg_reg],  # old values for inert regions
                    new_exp.gdf[chg_reg]        # new values for changing regions
                ]).loc[exposures.gdf.index,:],  # re-establish old order
            ),
            crs=exposures.crs
        )

        # set missing values of centr_
        if INDICATOR_CENTR + self.haz_type in new_exp.gdf.columns \
            and np.isnan(new_exp.gdf[INDICATOR_CENTR + self.haz_type].values).any():
            new_exp.gdf.drop(columns=INDICATOR_CENTR + self.haz_type, inplace=True)
        elif INDICATOR_CENTR in new_exp.gdf.columns \
            and np.isnan(new_exp.gdf[INDICATOR_CENTR].values).any():
            new_exp.gdf.drop(columns=INDICATOR_CENTR, inplace=True)

        # put hazard intensities outside region to previous intensities
        if hazard is not new_haz:
            if INDICATOR_CENTR + self.haz_type in exposures.gdf.columns:
                centr = exposures.gdf[INDICATOR_CENTR + self.haz_type].values[chg_reg]
            elif INDICATOR_CENTR in exposures.gdf.columns:
                centr = exposures.gdf[INDICATOR_CENTR].values[chg_reg]
            else:
                exposures.assign_centroids(hazard)
                centr = exposures.gdf[INDICATOR_CENTR + self.haz_type].values[chg_reg]

            centr = np.delete(np.arange(hazard.intensity.shape[1]), np.unique(centr))
            new_haz_inten = new_haz.intensity.tolil()
            new_haz_inten[:, centr] = hazard.intensity[:, centr]
            new_haz.intensity = new_haz_inten.tocsr()

        return new_exp, new_impfs, new_haz



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test select module.
"""


import unittest
import numpy as np

from climada.util.select import get_attributes_with_matching_dimension

class Dummy():

    def __init__(self):
        self.oneD3 = [1, 2, 3]
        self.oneD4 = [1, 2, 3, 4]
        self.twoD2 = [[1, 2, 3], [1, 2, 3, 4]]
        self.twoD3 = [[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]]
        self.twoD4 = [[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3 ,4]]
        self.twonp = np.array([[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]])

class TestGetAttributesDimension_pass(unittest.TestCase):
    """Test get_attributes_with_matching_dimension function"""

    def test_select_pass(self):
        """test get_attributes_with_matchin_dimension pass"""

        dummy = Dummy()
        list_attrs = get_attributes_with_matching_dimension(dummy, [3])
        self.assertTrue(np.array_equal(list_attrs, ['oneD3', 'twoD3', 'twonp']))

        list_attrs = get_attributes_with_matching_dimension(dummy, [4, 4])
        self.assertTrue(np.array_equal(list_attrs, ['twoD4']))
        list_attrs = get_attributes_with_matching_dimension(dummy, [3, 4])
        self.assertTrue(np.array_equal(list_attrs, ['twoD3', 'twonp']))

        list_attrs = get_attributes_with_matching_dimension(dummy, [5])
        self.assertTrue(np.array_equal(list_attrs, []))



# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestGetAttributesDimension_pass)
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define Uncertainty class.
"""

__all__ = ['UncOutput', 'UncCostBenefitOutput', 'UncImpactOutput']

import logging
import datetime as dt

from itertools import zip_longest
from pathlib import Path


import h5py
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl

from climada import CONFIG

from climada.util.value_representation import value_to_monetary_unit as u_vtm
from climada.util.value_representation import convert_monetary_value as u_cmv
from climada.util import plot as u_plot
import climada.util.hdf5_handler as u_hdf5

LOGGER = logging.getLogger(__name__)

# Metrics that are multi-dimensional
METRICS_2D = ['eai_exp', 'at_event']

DATA_DIR = CONFIG.engine.uncertainty.local_data.user_data.dir()

FIG_W, FIG_H = 8, 5 #default figize width/heigh column/work multiplicators

MAP_CMAP = 'Dark2' #Default color map for the sensitivity map

#Table of recommended pairing between salib sampling and sensitivity methods
# NEEDS TO BE UPDATED REGULARLY!! https://salib.readthedocs.io/en/latest/api.html
SALIB_COMPATIBILITY = {
    'fast': ['fast_sampler'],
    'rbd_fast': ['latin'] ,
    'morris': ['morris'],
    'sobol' : ['saltelli'],
    'delta' : ['latin'],
    'dgsm' : ['fast_sampler', 'latin', 'morris', 'saltelli', 'latin', 'ff'],
    'ff' : ['ff'],
    }

plt.style.use('seaborn-white')
params = {'legend.fontsize': 'x-large',
         'axes.labelsize': 'x-large',
         'axes.titlesize':'x-large',
         'xtick.labelsize':'x-large',
         'ytick.labelsize':'x-large'}
mpl.rcParams.update(params)


class UncOutput():
    """
    Class to store and plot uncertainty and sensitivity analysis output data

    This is the base class to store uncertainty and sensitivity outputs of an
    analysis  done on climada.engine.impact.Impact() or
    climada.engine.costbenefit.CostBenefit() object.

    Attributes
    ----------
    samples_df : pandas.DataFrame
        Values of the sampled uncertainty parameters. It has n_samples rows
        and one column per uncertainty parameter.
    sampling_method : str
        Name of the sampling method from SAlib.
        https://salib.readthedocs.io/en/latest/api.html#
    n_samples : int
        Effective number of samples (number of rows of samples_df)
    param_labels : list
        Name of all the uncertainty parameters
    distr_dict : dict
        Comon flattened dictionary of all the distr_dict of all input variables.
        It represents the distribution of all the uncertainty parameters.
    problem_sa : dict
        The description of the uncertainty variables and their
        distribution as used in SALib.
        https://salib.readthedocs.io/en/latest/basics.html.
    """

    _metadata = ['sampling_method', 'sampling_kwargs', 'sensitivity_method',
                 'sensitivity_kwargs']

    def __init__(self, samples_df, unit=None):
        """
        Initialize Uncertainty Data object.

        Parameters
        ----------
        samples_df : pandas.DataFrame
            input parameters samples
        unit : str, optional
            value unit
        """
        #Data
        self.samples_df = samples_df
        self.unit = unit

    def get_samples_df(self):
        return getattr(self, 'samples_df')

    def get_unc_df(self, metric_name):
        return getattr(self, f'{metric_name}_unc_df')

    def set_unc_df(self, metric_name, unc_df):
        setattr(self, f'{metric_name}_unc_df', unc_df)

    def get_sens_df(self, metric_name):
        return getattr(self, f'{metric_name}_sens_df')

    def set_sens_df(self, metric_name, sens_df):
        setattr(self, f'{metric_name}_sens_df', sens_df)

    def check_salib(self, sensitivity_method):
        """
        Checks whether the chosen sensitivity method and the sampling
        method used to generated self.samples_df
        respect the pairing recommendation by the SALib package.

        https://salib.readthedocs.io/en/latest/api.html

        Parameters
        ----------
        sensitivity_method : str
            Name of the sensitivity analysis method.

        Returns
        -------
        bool
            True if sampling and sensitivity methods respect the recommended
            pairing.

        """

        if self.sampling_method not in SALIB_COMPATIBILITY[sensitivity_method]:
            LOGGER.warning("The chosen combination of sensitivity method (%s)"
                " and sampling method (%s) does not correspond to the"
                " recommendation of the salib pacakge."
                "\n https://salib.readthedocs.io/en/latest/api.html",
                self.sampling_method, sensitivity_method
                )
            return False
        return True

    @property
    def sampling_method(self):
        """
        Returns the sampling method used to generate self.samples_df

        Returns
        -------
        str :
            Sampling method name

        """
        return self.samples_df.attrs['sampling_method']

    @property
    def sampling_kwargs(self):
        """
        Returns the kwargs of the sampling method that generate self.samples_df

        Returns
        -------
        dict
            Dictionary of arguments for SALib sampling method

        """
        return self.samples_df.attrs['sampling_kwargs']

    @property
    def n_samples(self):
        """
        The effective number of samples

        Returns
        -------
        int :
            effective number of samples

        """
        return self.samples_df.shape[0]

    @property
    def param_labels(self):
        """
        Labels of all uncertainty input parameters.

        Returns
        -------
        list of str :
            Labels of all uncertainty input parameters.

        """
        return list(self.samples_df)

    @property
    def problem_sa(self):
        """
        The description of the uncertainty variables and their
        distribution as used in SALib.
        https://salib.readthedocs.io/en/latest/basics.html

        Returns
        -------
        dict :
            Salib problem dictionary.

        """
        return {
            'num_vars' : len(self.param_labels),
            'names' : self.param_labels,
            'bounds' : [[0, 1]]*len(self.param_labels)
            }

    @property
    def uncertainty_metrics(self):
        """
        Retrieve all uncertainty output metrics names

        Returns
        -------
        unc_metric_list : [str]
            List of names of attributes containing metrics uncertainty values,
            without the trailing '_unc_df'

        """
        unc_metric_list = []
        for attr_name, attr_value in self.__dict__.items():
            if isinstance(attr_value, pd.DataFrame):
                if not attr_value.empty and attr_name.endswith("_unc_df"):
                    unc_metric_list.append(attr_name[:-7])
        return unc_metric_list

    @property
    def sensitivity_metrics(self):
        """
        Retrieve all sensitivity output metrics names

        Returns
        -------
        sens_metric_list : [str]
            List of names of attributes containing metrics sensitivity values,
            without the trailing '_sens_df'

        """
        sens_metric_list = []
        for attr_name, attr_value in self.__dict__.items():
            if isinstance(attr_value, pd.DataFrame):
                if not attr_value.empty and attr_name.endswith("_sens_df"):
                    sens_metric_list.append(attr_name[:-8])
        return sens_metric_list

    def get_uncertainty(self, metric_list=None):
        """
        Returns uncertainty dataframe with values for each sample

        Parameters
        ----------
        metric_list : [str], optional
            List of uncertainty metrics to consider.
            The default returns all uncertainty metrics at once.

        Returns
        -------
        pandas.DataFrame
            Joint dataframe of all uncertainty values for all metrics in
            the metric_list.

        See Also
        --------
        uncertainty_metrics: list of all available uncertainty metrics

        """
        if metric_list is None:
            metric_list = self.uncertainty_metrics
        try:
            unc_df = pd.concat(
                [self.get_unc_df(metric) for metric in metric_list],
                axis=1
                )
        except AttributeError:
            return pd.DataFrame([])
        return unc_df

    def get_sensitivity(self, salib_si, metric_list=None):
        """
        Returns sensitivity index

        E.g. For the sensitivity analysis method 'sobol', the choices
        are ['S1', 'ST'], for 'delta' the  choices are ['delta', 'S1'].

        For more information see the SAlib documentation:
        https://salib.readthedocs.io/en/latest/basics.html

        Parameters
        ----------
        salib_si : str
            Sensitivity index
        metric_list :[str], optional
            List of sensitivity metrics to consider.
            The default returns all sensitivity indices at once.

        Returns
        -------
        pandas.DataFrame
            Joint dataframe of the sensitvity indices for all metrics in
            the metric_list

        See Also
        --------
        sensitvity_metrics: list of all available sensitivity metrics

        """
        df_all = pd.DataFrame([])
        df_meta = pd.DataFrame([])

        if metric_list is None:
            metric_list = self.sensitivity_metrics
        for metric in metric_list:
            submetric_df = self.get_sens_df(metric)
            if not submetric_df.empty:
                submetric_df = submetric_df[submetric_df['si'] == salib_si]
                df_all = pd.concat(
                    [df_all, submetric_df.select_dtypes('number')],
                    axis=1
                    )
                if df_meta.empty:
                    df_meta = submetric_df.drop(
                        submetric_df.select_dtypes('number').columns, axis=1)
        return pd.concat([df_meta, df_all], axis=1).reset_index(drop=True)

    def get_largest_si(self, salib_si, metric_list=None, threshold=0.01):
        """
        Get largest si per metric

        Parameters
        ----------
        salib_si : str
            The name of the sensitivity index to plot.
        metric_list : list of strings, optional
            List of metrics to plot the sensitivity.
            Default is None.
        threshold : float
            The minimum value a sensitivity index must have to be considered
            as the largest. The default is 0.01.
        Returns
        -------
        max_si_df : pandas.dataframe
            Dataframe with the largest si and its value per metric

        """

        if metric_list is None:
            metric_list = self.sensitivity_metrics

        si_df = self.get_sensitivity(salib_si, metric_list)

        #get max index
        si_df_num = si_df.select_dtypes('number')
        si_df_num[si_df_num<threshold] = 0 #remove noise whenn all si are 0
        max_si_idx = si_df_num.idxmax().to_numpy()
        max_si_val = si_df_num.max().to_numpy()

        #get parameter of max index
        max_si_idx = np.nan_to_num(max_si_idx + 1).astype(int) # Set np.nan values to 0
        param = np.concatenate((['None'], si_df['param']))
        param_max_si = [param[idx] for idx in max_si_idx]
        param2 = np.concatenate((['None'], si_df['param2']))
        param2_max_si = [param2[idx] for idx in max_si_idx]

        max_si_df = pd.DataFrame(
            {'metric' : si_df_num.columns,
             'param' : param_max_si,
             'param2' : param2_max_si,
             'si' : max_si_val
             })

        return max_si_df

    def plot_sample(self, figsize=None):
        """
        Plot the sample distributions of the uncertainty input parameters

        For each uncertainty input variable, the sample distributions is shown
        in a separate axes.

        Parameters
        ---------
        figsize : tuple(int or float, int or float), optional
            The figsize argument of matplotlib.pyplot.subplots()
            The default is derived from the total number of plots (nplots) as:
                nrows, ncols = int(np.ceil(nplots / 3)), min(nplots, 3)
                figsize = (ncols * FIG_W, nrows * FIG_H)

        Raises
        ------
        ValueError
            If no sample was computed the plot cannot be made.

        Returns
        -------
        axes: matplotlib.pyplot.axes
            The axis handle of the plot.

        """
        fontsize = 18

        if self.samples_df.empty:
            raise ValueError("No uncertainty sample present."+
                    "Please make a sample first.")

        nplots = len(self.param_labels)
        nrows, ncols = int(np.ceil(nplots / 3)), min(nplots, 3)
        if not figsize:
            figsize = (ncols * FIG_W, nrows * FIG_H)
        _fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
        for ax, label in zip_longest(axes.flatten(),
                                     self.param_labels,
                                     fillvalue=None):
            if label is None:
                ax.remove()
                continue
            self.samples_df[label].hist(ax=ax, bins=100)
            ax.set_xlabel(label)
            ax.set_ylabel('Sample count')
            ax.tick_params(labelsize=fontsize)
            for item in [ax.title, ax.xaxis.label, ax.yaxis.label]:
                item.set_fontsize(fontsize)

        plt.tight_layout()

        return axes

    def plot_uncertainty(self, metric_list=None, orig_list=None, figsize=None,
                         log=False, axes=None):
        """
        Plot the  uncertainty distribution

        For each risk metric, a separate axes is used to plot the uncertainty
        distribution of the output values obtained over the sampled
        input parameters.

        Parameters
        ----------
        metric_list : list[str], optional
            List of metrics to plot the distribution.
            The default is None.
        orig_list : list[float], optional
            List of the original (without uncertainty) values for each
            sub-metric of the mtrics in metric_list. The ordering is identical.
            The default is None.
        figsize : tuple(int or float, int or float), optional
            The figsize argument of matplotlib.pyplot.subplots()
            The default is derived from the total number of plots (nplots) as:
            nrows, ncols = int(np.ceil(nplots / 3)), min(nplots, 3)
            figsize = (ncols * FIG_W, nrows * FIG_H)
        log : boolean, optional
            Use log10 scale for x axis. Default is False.
        axes : matplotlib.pyplot.axes, optional
            Axes handles to use for the plot. The default is None.

        Raises
        ------
        ValueError
            If no metric distribution was computed the plot cannot be made.

        Returns
        -------
        axes : matplotlib.pyplot.axes
            The axes handle of the plot.

        See Also
        --------
        uncertainty_metrics : list of all available uncertainty metrics

        """
        fontsize = 18 #default label fontsize

        if not self.uncertainty_metrics:
            raise ValueError("No uncertainty data present for these metrics. "+
                    "Please run an uncertainty analysis first.")

        if orig_list is None:
            orig_list = []

        if metric_list is None:
            metric_list = [
                metric
                for metric in self.uncertainty_metrics
                if metric not in METRICS_2D
                ]

        unc_df = self.get_uncertainty(metric_list)

        if unc_df.empty:
            raise AttributeError(f"The listed metrics {metric_list} either "
                 "do not exist or have no uncertainty values.")

        if log:
            unc_df_plt = unc_df.apply(np.log10).copy()
            unc_df_plt = unc_df_plt.replace([np.inf, -np.inf], np.nan)
        else:
            unc_df_plt = unc_df.copy()

        cols = unc_df_plt.columns
        nplots = len(cols)
        nrows, ncols = int(np.ceil(nplots / 2)), min(nplots, 2)
        if axes is None:
            if not figsize:
                figsize = (ncols * FIG_W, nrows * FIG_H)
            _fig, axes = plt.subplots(nrows = nrows,
                                     ncols = ncols,
                                     figsize = figsize)
        if nplots > 1:
            flat_axes = axes.flatten()
        else:
            flat_axes = np.array([axes])

        for ax, col, orig_val in zip_longest(flat_axes, cols, orig_list, fillvalue=None):
            if col is None:
                if ax is not None:
                    ax.remove()
                continue
            data, m_unit = u_vtm(unc_df_plt[col])
            data = pd.Series(data)
            if data.empty:
                ax.remove()
                continue
            data.hist(ax=ax, bins=30, density=True, histtype='bar',
                      color='lightsteelblue', edgecolor='black')
            try:
                data.plot.kde(ax=ax, color='darkblue', linewidth=4, label='')
            except np.linalg.LinAlgError:
                pass
            avg, std = data.mean(), data.std()
            _, ymax = ax.get_ylim()
            if log:
                avg_plot = np.log10(avg)
            else:
                avg_plot = avg
            ax.axvline(avg_plot, color='darkorange', linestyle='dashed', linewidth=2,
                    label="avg=%.2f%s" %(avg, m_unit))
            if orig_val is not None:
                if log:
                    orig_plot = np.log10(orig_val)
                else:
                    orig_plot = orig_val
                [orig_plot] = u_cmv(orig_plot, m_unit)
                ax.axvline(orig_plot, color='green', linestyle='dotted', linewidth=2,
                           label="orig=%.2f%s" %(orig_plot, m_unit))
            if log:
                std_m, std_p = np.log10(avg - std), np.log10(avg + std)
            else:
                std_m, std_p = avg - std, avg + std
            ax.plot([std_m, std_p],
                    [0.3 * ymax, 0.3 * ymax], color='black',
                    label="std=%.2f%s" %(std, m_unit))
            xlabel = col + ' [' + m_unit + ' ' + self.unit + '] '
            if log:
                ax.set_xlabel( xlabel + ' (log10 scale)')
            else:
                ax.set_xlabel(xlabel)
            ax.set_ylabel('density of samples')
            ax.legend(fontsize=fontsize-2)

            ax.tick_params(labelsize=fontsize)
            for item in [ax.title, ax.xaxis.label, ax.yaxis.label]:
                item.set_fontsize(fontsize)
            ax.set_xlim([data.min(), data.max()])
        plt.tight_layout()

        return axes


    def plot_rp_uncertainty(self, orig_list=None, figsize=(16, 6), axes=None):
        """
        Plot the distribution of return period uncertainty

        Parameters
        ----------
        orig_list : list[float], optional
            List of the original (without uncertainty) values for each
            sub-metric of the mtrics in metric_list. The ordering is identical.
            The default is None.
        figsize : tuple(int or float, int or float), optional
            The figsize argument of matplotlib.pyplot.subplots()
            The default is (8, 6)
        axes: matplotlib.pyplot.axes, optional
            Axes handles to use for the plot. The default is None.

        Raises
        ------
        ValueError
            If no metric distribution was computed the plot cannot be made.

        Returns
        -------
        ax : matplotlib.pyplot.axes
            The axis handle of the plot.

        """

        try:
            unc_df = self.freq_curve_unc_df
        except AttributeError:
            unc_df = None
        if unc_df is None or unc_df.empty:
            raise ValueError("No return period uncertainty data present "
                    "Please run an uncertainty analysis with the desired "
                    "return period specified.")

        add_orig=True
        if orig_list is None:
            add_orig=False

        if  axes is  None:
            _fig, axes = plt.subplots(figsize=figsize, nrows=1, ncols=2)

        [min_l, max_l], m_unit = u_vtm([unc_df.min().min(), unc_df.max().max()], n_sig_dig=4)

        ax = axes[0]

        prop_cycle = plt.rcParams['axes.prop_cycle']
        colors = prop_cycle.by_key()['color']

        for n, (_name, values) in enumerate(unc_df.iteritems()):
            values = u_cmv(values, m_unit, n_sig_dig=4)
            count, division = np.histogram(values, bins=100)
            count = count / count.max()
            losses = [(bin_i + bin_f )/2 for (bin_i, bin_f) in zip(division[:-1], division[1:])]
            ax.plot([min_l, max_l], [2*n, 2*n], color='k', alpha=0.5)
            ax.fill_between(losses, count + 2*n, 2*n)
            if add_orig:
                [orig_val] = u_cmv(orig_list[n], m_unit, n_sig_dig=4)
                ax.plot(
                    [orig_val, orig_val], [2*n, 2*(n+1)],
                    color=colors[n], linestyle='dotted', linewidth=2,
                    label="orig=%.2f%s" %(orig_val, m_unit)
                    )

        ax.set_xlim(min_l, max_l)
        ax.set_ylim(0, 2*unc_df.shape[1])
        ax.set_xlabel('Impact [%s %s]' %(m_unit, self.unit))
        ax.set_ylabel('Return period [years]')
        ax.set_yticks(np.arange(0, 2*unc_df.shape[1], 2))
        ax.set_yticklabels([s[2:] for s in unc_df.columns])
        ax.legend(loc='lower right')

        ax = axes[1]

        high = u_cmv(self.get_unc_df('freq_curve').quantile(0.95).values, m_unit, n_sig_dig=4)
        middle = u_cmv(self.get_unc_df('freq_curve').quantile(0.5).values, m_unit, n_sig_dig=4)
        low = u_cmv(self.get_unc_df('freq_curve').quantile(0.05).values, m_unit, n_sig_dig=4)

        x = [float(rp[2:]) for rp in unc_df.columns]
        ax.plot(x, high, linestyle='--', color = 'blue',
                alpha=0.5, label='0.95 percentile')
        ax.plot(x, middle, label='0.5 percentile')
        ax.plot(x, low, linestyle='dashdot', color='blue',
                alpha=0.5, label='0.05 percentile')
        ax.fill_between(x, low, high, alpha=0.2)
        if add_orig:
            ax.plot(x, u_cmv(orig_list, m_unit, n_sig_dig=4), color='green', linestyle='dotted', label='orig')
        ax.set_xlabel('Return period [year]')
        ax.set_ylabel('Impact [' + m_unit + ' ' + self.unit + ']')
        ax.legend()

        return axes


    def plot_sensitivity(self, salib_si='S1', salib_si_conf='S1_conf',
                         metric_list=None, figsize=None, axes=None,
                         **kwargs):
        """
        Bar plot of a first order sensitivity index

        For each metric, the sensitivity indices are plotted in a
        separate axes.

        This requires that a senstivity analysis was already
        performed.

        E.g. For the sensitivity analysis method 'sobol', the choices
        are ['S1', 'ST'], for 'delta' the  choices are ['delta', 'S1'].

        Note that not all sensitivity indices have a confidence interval.

        For more information see the SAlib documentation:
        https://salib.readthedocs.io/en/latest/basics.html

        Parameters
        ----------
        salib_si : string, optional
            The first order (one value per metric output) sensitivity index
            to plot.
            The default is S1.
        salib_si_conf : string, optional
            The  confidence value for the first order sensitivity index
            to plot.
            The default is S1_conf.
        metric_list : list of strings, optional
            List of metrics to plot the sensitivity. If a metric is not found
            it is ignored.
        figsize : tuple(int or float, int or float), optional
            The figsize argument of matplotlib.pyplot.subplots()
            The default is derived from the total number of plots (nplots) as:
                nrows, ncols = int(np.ceil(nplots / 3)), min(nplots, 3)
                figsize = (ncols * FIG_W, nrows * FIG_H)
        axes : matplotlib.pyplot.axes, optional
            Axes handles to use for the plot. The default is None.
        kwargs :
            Keyword arguments passed on to
            pandas.DataFrame.plot(kind='bar')

        Raises
        ------
        ValueError :
            If no sensitivity is available the plot cannot be made.

        Returns
        -------
        axes : matplotlib.pyplot.axes
            The axes handle of the plot.

        See Also
        --------
        sensitvity_metrics :
            list of all available sensitivity metrics

        """

        if not self.sensitivity_metrics:
            raise ValueError("No sensitivity present. "
                    "Please run a sensitivity analysis first.")

        if metric_list is None:
            metric_list = [
                metric
                for metric in self.sensitivity_metrics
                if metric not in METRICS_2D
                ]

        nplots = len(metric_list)
        nrows, ncols = int(np.ceil(nplots / 2)), min(nplots, 2)
        if axes is None:
            if not figsize:
                figsize = (ncols * FIG_W, nrows * FIG_H)
            _fig, axes = plt.subplots(nrows = nrows,
                                     ncols = ncols,
                                     figsize = figsize)
        if nplots > 1:
            flat_axes = axes.flatten()
        else:
            flat_axes = np.array([axes])

        for ax, metric in zip(flat_axes, metric_list):
            df_S = self.get_sensitivity(salib_si, [metric]).select_dtypes('number')
            if df_S.empty:
                ax.set_xlabel('Input parameter')
                ax.remove()
                continue
            df_S_conf = self.get_sensitivity(salib_si_conf, [metric]).select_dtypes('number')
            if df_S_conf.empty:
                df_S.plot(ax=ax, kind='bar', **kwargs)
            df_S.plot(ax=ax, kind='bar', yerr=df_S_conf, **kwargs)
            ax.set_xticklabels(self.param_labels, rotation=0)
            ax.set_xlabel('Input parameter')
            ax.set_ylabel(salib_si)
        plt.tight_layout()

        return axes

    def plot_sensitivity_second_order(self, salib_si='S2', salib_si_conf='S2_conf',
                                      metric_list=None, figsize=None, axes=None,
                                      **kwargs):
        """
        Plot second order sensitivity indices as matrix.

        For each metric, the sensitivity indices are plotted in a
        separate axes.

        E.g. For the sensitivity analysis method 'sobol', the choices
        are ['S2', 'S2_conf'].

        Note that not all sensitivity indices have a confidence interval.

        For more information see the SAlib documentation:
        https://salib.readthedocs.io/en/latest/basics.html

        Parameters
        ----------
        salib_si : string, optional
            The second order sensitivity indexto plot.
            The default is S2.
        salib_si_conf : string, optional
            The  confidence value for thesensitivity index salib_si
            to plot.
            The default is S2_conf.
        metric_list : list of strings, optional
            List of metrics to plot the sensitivity. If a metric is not found
            it is ignored. Default is all 1D metrics.
        figsize : tuple(int or float, int or float), optional
            The figsize argument of matplotlib.pyplot.subplots()
            The default is derived from the total number of plots (nplots) as:
                nrows, ncols = int(np.ceil(nplots / 3)), min(nplots, 3)
                figsize = (ncols * 5, nrows * 5)
        axes : matplotlib.pyplot.axes, optional
            Axes handles to use for the plot. The default is None.
        kwargs :
            Keyword arguments passed on to matplotlib.pyplot.imshow()

        Raises
        ------
        ValueError :
            If no sensitivity is available the plot cannot be made.

        Returns
        -------
        axes:  matplotlib.pyplot.axes
            The axes handle of the plot.

        See Also
        --------
        sensitvity_metrics :
            list of all available sensitivity metrics

        """

        if not self.sensitivity_metrics:
            raise ValueError("No sensitivity present for this metrics. "
                    "Please run a sensitivity analysis first.")

        if metric_list is None:
            metric_list = [
                metric
                for metric in self.sensitivity_metrics
                if metric not in METRICS_2D
                ]


        if 'cmap' not in kwargs.keys():
            kwargs['cmap'] = 'summer'

        #all the lowest level metrics (e.g. rp10) directly or as
        #submetrics of the metrics in metrics_list
        df_S = self.get_sensitivity(salib_si, metric_list).select_dtypes('number')
        df_S_conf = self.get_sensitivity(salib_si_conf, metric_list).select_dtypes('number')

        nplots = len(df_S.columns)
        nrows, ncols = int(np.ceil(nplots / 3)), min(nplots, 3)
        if axes is None:
            if not figsize:
                figsize = (ncols * 5, nrows * 5)
            _fig, axes = plt.subplots(nrows = nrows,
                                     ncols = ncols,
                                     figsize = figsize)

        if nplots > 1:
            flat_axes = axes.flatten()
        else:
            flat_axes = np.array([axes])

        for ax, submetric in zip(flat_axes, df_S.columns):
            #Make matrix symmetric
            s2_matrix = np.triu(
                np.reshape(
                    df_S[submetric].to_numpy(),
                    (len(self.param_labels), -1)
                    )
                )
            s2_matrix = s2_matrix + s2_matrix.T - np.diag(np.diag(s2_matrix))
            ax.imshow(s2_matrix, **kwargs)
            s2_conf_matrix = np.triu(
                np.reshape(
                    df_S_conf[submetric].to_numpy(),
                    (len(self.param_labels), -1)
                    )
                )
            s2_conf_matrix = s2_conf_matrix + s2_conf_matrix.T - \
                np.diag(np.diag(s2_conf_matrix))
            for i in range(len(s2_matrix)):
                for j in range(len(s2_matrix)):
                    if np.isnan(s2_matrix[i, j]):
                        ax.text(j, i, np.nan,
                           ha="center", va="center",
                           color="k", fontsize='medium')
                    else:
                        ax.text(j, i,
                            str(round(s2_matrix[i, j], 2)) + u'\n\u00B1' +  #\u00B1 = +-
                            str(round(s2_conf_matrix[i, j], 2)),
                            ha="center", va="center",
                            color="k", fontsize='medium')

            ax.set_title(salib_si + ' - ' + submetric, fontsize=18)
            labels = self.param_labels
            ax.set_xticks(np.arange(len(labels)))
            ax.set_yticks(np.arange(len(labels)))
            ax.set_xticklabels(labels, fontsize=16)
            ax.set_yticklabels(labels, fontsize=16)
        plt.tight_layout()

        return axes

    def plot_sensitivity_map(self, salib_si='S1', **kwargs):
        """
        Plot a map of the largest sensitivity index in each exposure point

        Requires the uncertainty distribution for eai_exp.

        Parameters
        ----------
        salib_si : str, optional
            The name of the sensitivity index to plot.
            The default is 'S1'.
        kwargs :
            Keyword arguments passed on to
            climada.util.plot.geo_scatter_categorical

        Raises
        ------
        ValueError :
            If no sensitivity data is found, raise error.

        Returns
        -------
        ax: matplotlib.pyplot.axes
            The axis handle of the plot.

        See Also
        --------
        climada.util.plot.geo_scatter_categorical :
            geographical plot for categorical variable

        """

        eai_max_si_df = self.get_largest_si(salib_si, metric_list=['eai_exp'])

        plot_val = eai_max_si_df['param']
        coord = np.array([self.coord_df.latitude, self.coord_df.longitude]).transpose()
        if 'var_name' not in kwargs:
            kwargs['var_name'] = 'Input parameter with largest ' + salib_si
        if 'title' not in kwargs:
            kwargs['title'] = ''
        if 'figsize' not in kwargs:
            kwargs['figsize'] = (8,6)
        if 'cmap' not in kwargs:
            labels = np.unique(plot_val)
            n=np.where(labels=='None')[0]
            if len(n) > 0 :
                n = n[0]
                cmap = mpl.colors.ListedColormap(
                    plt.get_cmap(MAP_CMAP).colors[:len(labels)]
                    )
                colors = list(cmap.colors)
                colors[n] = tuple(np.repeat(0.93, 3))
                cmap.colors = tuple(colors)
                kwargs['cmap'] = cmap
        ax = u_plot.geo_scatter_categorical(
                plot_val, coord,
                **kwargs
                )

        return ax

    def to_hdf5(self, filename=None):
        """
        Save output to .hdf5

        Parameters
        ----------
        filename : str or pathlib.Path, optional
            The filename with absolute or relative path.
            The default name is "unc_output + datetime.now() + .hdf5" and
            the default path is taken from climada.config

        Returns
        -------
        save_path : pathlib.Path
            Path to the saved file

        """
        if filename is None:
            filename = "unc_output" + dt.datetime.now().strftime(
                                                            "%Y-%m-%d-%H%M%S"
                                                            )
            filename = Path(DATA_DIR) / Path(filename)
        save_path = Path(filename)
        save_path = save_path.with_suffix('.hdf5')

        LOGGER.info('Writing %s', save_path)
        store = pd.HDFStore(save_path)
        for (var_name, var_val) in self.__dict__.items():
            if isinstance(var_val, pd.DataFrame):
                store.put(var_name, var_val, format='fixed', complevel=9)
        store.get_storer('/samples_df').attrs.metadata = self.samples_df.attrs
        store.close()

        str_dt = h5py.special_dtype(vlen=str)
        with h5py.File(save_path, 'a') as fh:
            if getattr(self, 'unit'):
                fh['impact_unit'] = [self.unit]
            if hasattr(self, 'sensitivity_method'):
                if self.sensitivity_method:
                    fh['sensitivity_method'] = [self.sensitivity_method]
            if hasattr(self, 'sensitivity_kwargs'):
                if self.sensitivity_kwargs:
                    grp = fh.create_group("sensitivity_kwargs")
                    for key, value in dict(self.sensitivity_kwargs).items():
                        ds = grp.create_dataset(key, (1,), dtype=str_dt)
                        ds[0] = str(value)
        return save_path

    @staticmethod
    def from_hdf5(filename):
        """
        Load a uncertainty and uncertainty output data from .hdf5 file

        Parameters
        ----------
        filename : str or pathlib.Path
            The filename with absolute or relative path.

        Returns
        -------
        unc_output: climada.engine.uncertainty.unc_output.UncOutput
            Uncertainty and sensitivity data loaded from .hdf5 file.
        """
        filename = Path(filename).with_suffix('.hdf5')
        if not filename.exists():
            LOGGER.info('File not found')
            return None

        unc_data = UncOutput(pd.DataFrame())

        LOGGER.info('Reading %s', filename)
        store = pd.HDFStore(filename)
        for var_name in store.keys():
            setattr(unc_data, var_name[1:], store.get(var_name))
        unc_data.samples_df.attrs = store.get_storer('/samples_df').attrs.metadata
        store.close()
        with h5py.File(filename, 'r') as fh:
            if 'impact_unit' in list(fh.keys()):
                unc_data.unit = fh.get('impact_unit')[0].decode('UTF-8')
            if 'sensitivity_method' in list(fh.keys()):
                unc_data.sensitivity_method = \
                    fh.get('sensitivity_method')[0].decode('UTF-8')
            if 'sensitivity_kwargs' in list(fh.keys()):
                grp = fh["sensitivity_kwargs"]
                sens_kwargs = {
                    key: u_hdf5.to_string(grp.get(key)[0])
                    for key in grp.keys()
                    }
                unc_data.sensitivity_kwargs = tuple(sens_kwargs.items())
        return unc_data


class UncImpactOutput(UncOutput):
    """Extension of UncOutput specific for CalcImpact, returned by the  uncertainty() method.
    """
    def __init__(self, samples_df, unit, aai_agg_unc_df, freq_curve_unc_df, eai_exp_unc_df,
                 at_event_unc_df, tot_value_unc_df, coord_df):
        """Constructor

        Uncertainty output values from impact.calc for each sample

        Parameters
        ----------
        samples_df : pandas.DataFrame
            input parameters samples
        unit : str
            value unit
        aai_agg_unc_df : pandas.DataFrame
            Each row contains the value of aai_aag for one sample (row of
            samples_df)
        freq_curve_unc_df : pandas.DataFrame
            Each row contains the values of the impact exceedence frequency
            curve for one sample (row of samples_df)
        eai_exp_unc_df : pandas.DataFrame
            Each row contains the values of eai_exp for one sample (row of
            samples_df)
        at_event_unc_df : pandas.DataFrame
            Each row contains the values of at_event for one sample (row of
            samples_df)
        tot_value_unc_df : pandas.DataFrame
            Each row contains the value of tot_value for one sample (row of
            samples_df)
        coord_df : pandas.DataFrame
            Coordinates of the exposure
        """
        super().__init__(samples_df, unit)
        self.aai_agg_unc_df = aai_agg_unc_df
        self.aai_agg_sens_df = None
        self.freq_curve_unc_df = freq_curve_unc_df
        self.freq_curve_sens_df = None
        self.eai_exp_unc_df = eai_exp_unc_df
        self.eai_exp_sens_df = None
        self.at_event_unc_df = at_event_unc_df
        self.at_event_sens_df = None
        self.tot_value_unc_df = tot_value_unc_df
        self.tot_value_sens_df = None
        self.coord_df = coord_df


class UncCostBenefitOutput(UncOutput):
    """Extension of UncOutput specific for CalcCostBenefit, returned by the uncertainty() method.
    """
    def __init__(self, samples_df, unit, imp_meas_present_unc_df, imp_meas_future_unc_df,
                 tot_climate_risk_unc_df, benefit_unc_df, cost_ben_ratio_unc_df,
                 cost_benefit_kwargs):
        """Constructor

        Uncertainty output values from cost_benefit.calc for each sample

        Parameters
        ----------
        samples_df : pandas.DataFrame
            input parameters samples
        unit : str
            value unit
        imp_meas_present_unc_df : pandas.DataFrame
            Each row contains the values of imp_meas_present for one sample
            (row of samples_df)
        imp_meas_future_unc_df : pandas.DataFrame
            Each row contains the values of imp_meas_future for one sample
            (row of samples_df)
        tot_climate_risk_unc_df : pandas.DataFrame
            Each row contains the values of tot_cliamte_risk for one sample
            (row of samples_df)
        benefit_unc_df : pandas.DataFrame
            Each row contains the values of benefit for one sample
            (row of samples_df)
        cost_ben_ratio_unc_df : pandas.DataFrame
            Each row contains the values of cost_ben_ratio for one sample
            (row of samples_df)
        cost_benefit_kwargs : pandas.DataFrame
            Each row contains the value of cost_benefit for one sample
            (row of samples_df)

        """
        super().__init__(samples_df, unit)
        self.imp_meas_present_unc_df= imp_meas_present_unc_df
        self.imp_meas_present_sens_df = None
        self.imp_meas_future_unc_df= imp_meas_future_unc_df
        self.imp_meas_future_sens_df = None
        self.tot_climate_risk_unc_df = tot_climate_risk_unc_df
        self.tot_climate_risk_sens_df = None
        self.benefit_unc_df = benefit_unc_df
        self.benefit_sens_df = None
        self.cost_ben_ratio_unc_df = cost_ben_ratio_unc_df
        self.cost_ben_ratio_sens_df = None
        self.cost_benefit_kwargs = cost_benefit_kwargs



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test config module.
"""
import unittest
import logging

from climada.util import log_level

class TestUtilInit(unittest.TestCase):
    """Test util __init__ methods"""

    def test_log_level_pass(self):
        """Test log level context manager passes"""
        #Check loggers are set to level
        with self.assertLogs('climada', level='INFO') as cm:
             with log_level('WARNING'):
                logging.getLogger('climada').info('info')
                logging.getLogger('climada').error('error')
                self.assertEqual(cm.output, ['ERROR:climada:error'])
        #Check if only climada loggers level change
        with self.assertLogs('matplotlib', level='DEBUG') as cm:
            with log_level('ERROR', name_prefix='climada'):
                logging.getLogger('climada').info('info')
            logging.getLogger('matplotlib').debug('debug')
            self.assertEqual(cm.output, ['DEBUG:matplotlib:debug'])

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestUtilInit)
    unittest.TextTestRunner(verbosity=2).run(TESTS)
